"""Respond to an action from the user on the NetGraph"""

from nengo.utils.compat import iteritems

import nengo_gui.components


def create_action(action, net_graph, **kwargs):
    if action == "expand":
        return ExpandCollapse(net_graph, expand=True, **kwargs)
    elif action == "collapse":
        return ExpandCollapse(net_graph, expand=False, **kwargs)
    elif action == "create_graph":
        return CreateGraph(net_graph, **kwargs)
    elif action == "pos":
        return Pos(net_graph, **kwargs)
    elif action == "size":
        return Size(net_graph, **kwargs)
    elif action == "pos_size":
        return PosSize(net_graph, **kwargs)
    elif action == "feedforward_layout":
        return FeedforwardLayout(net_graph, **kwargs)
    elif action == "config":
        return ConfigAction(net_graph, **kwargs)
    else:
        return Action(net_graph, **kwargs)


class Action(object):
    """Base object for all user action responses"""

    def __init__(self, net_graph, uid):
        self.net_graph = net_graph
        self.uid = uid

    @property
    def obj(self):
        if self.uid is None:
            raise ValueError("Cannot get 'obj' without 'uid'")
        return self.net_graph.uids[self.uid]

    @property
    def obj_config(self):
        if self.uid is None:
            raise ValueError("Cannot get 'obj' without 'uid'")
        return self.net_graph.page.config[self.obj]

    def send(self, action, **kwargs):
        if "uid" not in kwargs:
            kwargs["uid"] = self.uid
        kwargs["type"] = action
        self.net_graph.to_be_sent.append(kwargs)

    def apply(self):
        raise NotImplementedError('Subclasses should implement this.')

    def undo(self):
        raise NotImplementedError('Subclasses should implement this.')


class ConfigAction(Action):
    """Update config file entry related to a visualisation component"""

    def __init__(self, page, component, new_cfg, old_cfg):
        super(ConfigAction, self).__init__(page.net_graph, id(component))
        self.component = component
        self.page = page
        self.new_cfg = new_cfg
        self.old_cfg = old_cfg

    def load(self, cfg):
        for k, v in iteritems(cfg):
            setattr(self.page.config[self.component], k, v)
        self.net_graph.modified_config()
        self.send("config", config=cfg)

    def apply(self):
        self.load(self.new_cfg)

    def undo(self):
        self.load(self.old_cfg)


class ExpandCollapse(Action):
    """Expand and collapse the Network NetGraph item"""

    def __init__(self, net_graph, uid, expand):
        super(ExpandCollapse, self).__init__(net_graph, uid)
        self.expand = expand
        if expand:
            self.net_graph.act_expand(self.uid)
        else:
            self.net_graph.act_collapse(self.uid)

    def apply(self):
        if self.expand:
            self.send('expand')
        else:
            self.send('collapse')

    def undo(self):
        if self.expand:
            self.send('collapse')
        else:
            self.send('expand')


class RemoveGraph(Action):
    """Remove a visualisation component associated to a NetGraph item"""

    def __init__(self, net_graph, component):
        super(RemoveGraph, self).__init__(net_graph, id(component))
        self.component = component

    def apply(self):
        self.send('delete_graph')

    def undo(self):
        page = self.net_graph.page
        page.add_component(self.component)

        page.locals[self.component.uid] = self.component
        page.default_labels[self.component] = self.component.uid

        page.changed = True
        self.send('js', code=self.component.javascript())


class CreateGraph(Action):
    """Create a visualisation component associated to a NetGraph item"""

    def __init__(self, net_graph, uid, type, x, y, width, height, **kwargs):
        super(CreateGraph, self).__init__(net_graph, uid)
        self.graph_uid = None
        self.type = type
        self.x, self.y = x, y
        self.width, self.height = width, height
        cls = getattr(nengo_gui.components, self.type)
        self.component = cls(self.obj, **kwargs)

        # If only one instance of the component is allowed, and another had to be
        # destroyed to create this one, keep track of it here so it can be undone
        self.duplicate = None

        # Remove any existing sliders associated with the same node
        if type == 'Slider':
            for component in self.net_graph.page.components:
                if (isinstance(component, nengo_gui.components.slider.Slider)
                        and component.node is self.obj):
                    self.duplicate = RemoveGraph(net_graph, component)
                    self.send('delete_graph', uid=component.original_id)

        self.act_create_graph()

    def act_create_graph(self):
        if self.graph_uid is None:
            self.net_graph.page.generate_uid(self.component, prefix='_viz_')
            self.graph_uid = self.net_graph.page.get_uid(self.component)
        else:
            self.net_graph.page.locals[self.graph_uid] = self.component
            self.net_graph.page.default_labels[self.component] = (
                self.graph_uid)
        self.net_graph.page.config[self.component].x = self.x
        self.net_graph.page.config[self.component].y = self.y
        self.net_graph.page.config[self.component].width = self.width
        self.net_graph.page.config[self.component].height = self.height
        self.net_graph.modified_config()

        self.net_graph.page.add_component(self.component)
        self.net_graph.page.changed = True
        self.send('js', code=self.component.javascript())

    def apply(self):
        if self.duplicate is not None:
            self.duplicate.apply()
        self.act_create_graph()

    def undo(self):
        self.send('delete_graph', uid=self.component.original_id)
        if self.duplicate is not None:
            self.duplicate.undo()


class PosSize(Action):
    """Set size and position on a NetGraph Item or a visualisation component"""

    def __init__(self, net_graph, uid, x, y, width, height):
        super(PosSize, self).__init__(net_graph, uid)
        self.x, self.y = self.obj_config.pos
        self.width, self.height = self.obj_config.size

        self.act_pos_size(x, y, width, height)

    def act_pos_size(self, x, y, width, height):
        self.obj_config.pos = x, y
        self.obj_config.size = width, height
        self.net_graph.modified_config()

    def apply(self):
        x, y, width, height = self.x, self.y, self.width, self.height
        self.x, self.y = self.obj_config.pos
        self.width, self.height = self.obj_config.size
        self.act_pos_size(x, y, width, height)
        self.send('pos_size', pos=[x, y], size=[width, height])

    def undo(self):
        self.apply()  # PosSize is a mirrored operation


class Pos(Action):
    """Set the position on a NetGraph Item or a visualisation component"""

    def __init__(self, net_graph, uid, x, y):
        super(Pos, self).__init__(net_graph, uid)
        self.x, self.y = self.obj_config.pos

        self.act_pos(x, y)

    def act_pos(self, x, y):
        self.obj_config.pos = x, y
        self.net_graph.modified_config()

    def apply(self):
        x, y = self.x, self.y
        self.x, self.y = self.obj_config.pos
        width, height = self.obj_config.size
        self.act_pos(x, y)
        self.send('pos_size', pos=[x, y], size=[width, height])

    def undo(self):
        self.apply()  # Pos is a mirrored operation


class Size(Action):
    """Set the size on a NetGraph Item or a visualisation component"""

    def __init__(self, net_graph, uid, width, height):
        super(Size, self).__init__(net_graph, uid)
        self.width, self.height = self.obj_config.size

        self.act_size(width, height)

    def act_size(self, width, height):
        self.obj_config.size = width, height
        self.net_graph.modified_config()

    def apply(self):
        width, height = self.width, self.height
        x, y = self.obj_config.pos
        self.width, self.height = self.obj_config.size
        self.act_size(width, height)
        self.send('pos_size', pos=[x, y], size=[width, height])

    def undo(self):
        self.apply()  # Size is a mirrored operation


class FeedforwardLayout(Action):
    """React to the auto-layout command"""

    def __init__(self, net_graph, uid):
        super(FeedforwardLayout, self).__init__(net_graph, uid)

        if self.uid is None:
            self.network = self.net_graph.page.model
            self.scale = self.net_graph.page.config[self.network].size[0]
            self.x, self.y = self.net_graph.page.config[self.network].pos
        else:
            self.network = self.obj
            self.scale = 1.0
            self.x, self.y = 0, 0

        self.pos = self.net_graph.layout.make_layout(self.network)
        # record the current positions and sizes of everything in the network
        self.old_state = self.save_network()
        self.act_feedforward_layout()
        self.new_state = self.save_network()

    def act_feedforward_layout(self):
        for obj, layout in iteritems(self.pos):
            obj_cfg = self.net_graph.page.config[obj]
            obj_cfg.pos = (layout['y'] / self.scale - self.x,
                           layout['x'] / self.scale - self.y)
            obj_cfg.size = (layout['h'] / 2 / self.scale,
                            layout['w'] / 2 / self.scale)

            obj_uid = self.net_graph.page.get_uid(obj)

            self.send('pos_size',
                      uid=obj_uid, pos=obj_cfg.pos, size=obj_cfg.size)

        self.net_graph.page.config[self.network].has_layout = True
        self.net_graph.modified_config()

    def save_network(self):
        state = []
        for obj, layout in iteritems(self.pos):
            state.append({
                'uid': self.net_graph.page.get_uid(obj),
                'pos': self.net_graph.page.config[obj].pos,
                'size': self.net_graph.page.config[obj].size,
                'obj': obj,
            })
        return state

    def load_network(self, state):
        for item in state:
            self.send('pos_size',
                      uid=item['uid'], pos=item['pos'], size=item['size'])
            self.net_graph.page.config[item['obj']].pos = item['pos']
            self.net_graph.page.config[item['obj']].size = item['size']
        # TODO: should config[network].has_layout be changed here?
        self.net_graph.modified_config()

    def apply(self):
        self.load_network(self.new_state)

    def undo(self):
        self.load_network(self.old_state)



import contextlib
import importlib
import os
import threading
import traceback
import sys
from nengo.utils.compat import StringIO


# list of Simulators to check for
known_modules = ['nengo', 'nengo_ocl', 'nengo_distilled',
                 'nengo_dl', 'nengo_mpi',
                 'nengo_brainstorm', 'nengo_spinnaker']


def discover_backends():
    found_modules = {}
    for name in known_modules:
        try:
            mod = importlib.import_module(name)
        except Exception as e:
            # TODO only ignore ImportErrors "No module named ...", display
            # other errors to the user as they might help debugging broken
            # backend installations
            continue
        found_modules[name] = mod
    return found_modules


class StartedSimulatorException(Exception):
    pass


class StartedGUIException(Exception):
    pass


# create a wrapper class that will throw an exception if we are
# currently executing a script
def make_dummy(cls):
    class DummySimulator(cls):
        def __init__(self, *args, **kwargs):
            if is_executing():
                raise StartedSimulatorException()
            super(DummySimulator, self).__init__(*args, **kwargs)
    return DummySimulator


# thread local storage for storing whether we are executing a script
flag = threading.local()

compiled_filename = '<nengo_gui_compiled>'

def is_executing():
    return getattr(flag, 'executing', False)


def determine_line_number():
    '''Checks stack trace to determine the line number we are currently at.

    The filename argument should be the filename given to the code when
    it was compiled (with compile())
    '''

    exc_type, exc_value, exc_traceback = sys.exc_info()
    if exc_traceback is not None:
        ex_tb = traceback.extract_tb(exc_traceback)
        for fn, line, function, code in reversed(ex_tb):
            if fn == compiled_filename:
                return line

    # if we can't find it that way, parse the text of the stack trace
    #  note that this is required for indentation errors and other syntax
    #  problems
    trace = traceback.format_exc()
    pattern = 'File "%s", line ' % compiled_filename
    index = trace.find(pattern)
    if index >=0:
        text = trace[index + len(pattern):].split('\n', 1)[0]
        if ',' in text:
            text = text.split(',', 1)[0]
        line = int(text)
        return line
    return None


class ExecutionEnvironment(object):
    def __init__(self, filename, allow_sim=False):
        if filename is None:
            self.directory = None
        else:
            self.directory = os.path.dirname(filename)
        self.added_directory = None
        self.allow_sim = allow_sim
    def __enter__(self):
        if self.directory is not None and self.directory not in sys.path:
            sys.path.insert(0, self.directory)
            self.added_directory = self.directory

        self.stdout = StringIO()

        flag.executing = True
        self.simulators = {}

        # add hooks to record stdout

        sys.stdout = self.stdout

        if not self.allow_sim:
            for mod in discover_backends().values():
                self.simulators[mod] = mod.Simulator
                mod.Simulator = make_dummy(mod.Simulator)

    def __exit__(self, exc_type, exc_value, traceback):
        for mod, cls in self.simulators.items():
            mod.Simulator = cls
        flag.executing = False

        sys.stdout = sys.__stdout__

        # ensure what has been printed is safe to show in html
        s = self.stdout.getvalue()
        s = s.replace('<', '&lt;').replace('>', '&gt;')
        self.stdout = StringIO(s)

        if not self.allow_sim:
            if self.added_directory is not None:
                sys.path.remove(self.added_directory)
                self.added_directory = None



# Routed Sequencing

# This model introduces routing in the sequencing model. The addition of routing
# allows the system to choose between two different actions: whether to go 
# through the sequence, or be driven by the visual input as explained in the 
# book. For instance, if the visual input has its value set to 0.8*START+D, the 
# model will begin cycling through at D->E, etc. Thus in this model, the input 
# doesn't prevent the activation of the second rule in the sequence.

# The parameters used in the model are as described in the book, with 16 
# dimensions for all semantic pointers.

# In Nengo 1.4, a buffer element for representing the vision was created by 
# using Buffer() as described in the book. However, in Nengo 2.0, you will have 
# to use State() with feedback parameter set to 0 (which is the default value in
# nengo).

# Press the play button in the visualizer to run the simulation. 
# The graph on the bottom-left shows the visual input recieved by the model, 
# the state graph in the middle shows the semantic pointer representation of the
# values stored in the state ensemble. The actions graph on bottom-right shows 
# the current transition or the action being executed, and the state graph on 
# top-right shows the utility (similarity) of the current basal ganglia input 
# (i.e., state) with the possible vocabulary vectors.

# You can see that in this model, even though the input is applied for 400ms, it 
# doesn't prevent the activation of the second and subsequent rules in the 
# sequence.

#Setup the environment
import nengo
from nengo import spa      #import spa related packages

#Number of dimensions for the Semantic Pointers
dimensions = 16

#Make a model object with the SPA network
model = spa.SPA(label='Routed_Sequence', seed=20)

with model:
    #Specify the modules to be used
    model.state = spa.State(dimensions=dimensions, feedback=1, 
                                    feedback_synapse=0.01)
    model.vision = spa.State(dimensions=dimensions) 
    
    #Specify the action mapping
    actions = spa.Actions(
        'dot(vision, START) --> state = vision',
        'dot(state, A) --> state = B',
        'dot(state, B) --> state = C',
        'dot(state, C) --> state = D',
        'dot(state, D) --> state = E',
        'dot(state, E) --> state = A'
    )
    
    #Creating the BG and Thalamus components that confirm to the specified rules
    model.BG = spa.BasalGanglia(actions=actions)
    model.thal = spa.Thalamus(model.BG)
    
    #Function that provides the model with an initial input semantic pointer.
    def start(t):
        if t < 0.4:
            return '0.8*START+D'
        else:
            return '0'

    #Input
    model.input = spa.Input(vision=start)


# Nengo Example: Integrator
#
# This demo implements a one-dimensional neural integrator.
#
# This is the first example of a recurrent network in the demos. It shows how
# neurons can be used to implement stable dynamics. Such dynamics are 
# important for memory, noise cleanup, statistical inference, and many 
# other dynamic transformations.
#
# Note that since the integrator constantly sums its input, it will 
# saturate quickly if you leave the input non-zero. This makes it clear 
# that neurons have a finite range of representation. Such saturation 
# effects can be exploited to perform useful computations
# (e.g. soft normalization).

import nengo

model = nengo.Network()
with model:
    # Our ensemble consists of 100 leaky integrate-and-fire neurons,
    # representing a one-dimensional signal
    a = nengo.Ensemble(n_neurons=100, dimensions=1)

    # Create a piecewise step function for input
    stim = nengo.Node([0])
    
    # Connect the population to itself using a long time constant (tau) 
    # for stability
    tau = 0.1
    nengo.Connection(a, a, synapse=tau)

    # Connect the input using the same time constant as on the recurrent
    # connection to make it more ideal
    nengo.Connection(stim, a, transform=tau, synapse=tau)


# Tutorial 3: Many neurons

# Brains have many neurons.  What happens if we use 20 neurons? 50? 100?

# The code below shows 20 neurons in a group (we call groups of neurons
# an "Ensemble" of neurons).  The representation is much more accurate.
# To see this, press Play.  Now, as you move the slider, the bottom graph
# should follow your movements well.  

# Try changing the number of neurons by editing the code below where it says
# "n_neurons=20".  Try 50.  Try 100.  The representation should get more and
# more accurate.

# Nengo Tip: Don't use too big a number!  Depending on your computer, using
# more than 1000 neurons in a single Ensemble can take a long time to compute.
# This is because Nengo needs to figure out how to weight the outputs of all
# the individual neurons (what we call "decoders"), in order to get the bottom
# graph.

import nengo

model = nengo.Network()
with model:
    stim = nengo.Node(0)
    ens = nengo.Ensemble(n_neurons=20, dimensions=1)
    nengo.Connection(stim, ens)



import os
import time

import pytest

from nengo_gui import testing_tools as tt

test_files = tt.folder_location('examples/tutorial')
ignore = ["spa", "networks"]
test_files = [f for f in test_files if not any(i in f for i in ignore)]
# Gets rid of SPA and Network files


@pytest.mark.parametrize('test_file', test_files)
def test_tutorial_basic(driver, test_file):
    # Tests the first 18 tutorials. These are the tutorials that do not
    # utilize networks or SPA which require a different format of test

    try:
        # Test page response by clicking the reset button and applying
        # new code to ace-editor
        tt.reset_page(driver)
        time.sleep(1)
        tt.update_editor(driver, test_file)
        tt.mouse_scroll(driver, 500)
        time.sleep(2)
        side_script = ("var right = document.getElementById('rightpane');\n"
                       "right.style.width = '200px';\n")
        driver.execute_script(side_script)

        node_objects = driver.find_elements_by_xpath('//*[@class="node"]')
        ens_objects = driver.find_elements_by_xpath('//*[@class="ens"]')

        node_number = test_file.count("nengo.Node")
        ens_number = test_file.count("nengo.Ensemble")

        # Makes sure the correct number of ensembles and nodes were rendered
        assert(len(node_objects)/2 == node_number)
        assert(len(ens_objects)/2 == ens_number)

        # Tests whether the file compiles and runs
        hang_time = 25  # alloted time until test fails

        time_script = ("var time = $('#ticks_tr');\n"
                       "return time.find('td').text()\n")
        tt.start_stop_sim(driver)
        time_start = time.time()
        while time.time() - time_start < hang_time:
            sim_time = driver.execute_script(time_script)
            if float(sim_time) > 0:
                break
            time.sleep(0.1)
        else:
            assert False, "Did not compile in %f seconds" % hang_time

    except:
        # Travis Only: On fail takes screenshot and uploads it to imgur
        if('TRAVIS' in os.environ):
            tt.imgur_screenshot(driver)
        raise



import re
try:
    from io import BytesIO
except ImportError:  # Python 2.7
    from BytesIO import BytesIO

from nengo_gui import server


class TwoWayStringIO(object):
    def __init__(self, readable_data):
        self.readable_data = readable_data
        self.read_buffer = BytesIO(self.readable_data)
        self.write_buffer = BytesIO()
        self.written_data = None
        self.closed = False

    def read(self, *args, **kwargs):
        return self.read_buffer.read(*args, **kwargs)

    def readline(self, *args, **kwargs):
        return self.read_buffer.readline(*args, **kwargs)

    def write(self, *args, **kwargs):
        return self.write_buffer.write(*args, **kwargs)

    def close(self):
        self.closed = True
        self.written_data = self.write_buffer.getvalue()
        self.read_buffer.close()

    def flush(self):
        self.write_buffer.flush()


class ServerMock(object):
    server_name = 'localhost'
    server_port = 80

    def create_websocket(self, socket):
        return server.WebSocket(socket)


class SocketMock(object):
    def __init__(self, request_data):
        self.request_data = request_data.encode('utf-8')
        self.file_io = TwoWayStringIO(self.request_data)

    def makefile(self, mode='r', bufsize=1024):
        return self.file_io

    def send(self, data):
        self.file_io.write(data)
        return len(data)

    def sendall(self, data):
        self.send(data)

    def setblocking(self, value):
        pass


class TestHttpWsRequestHandler(object):
    def test_get(self):
        request = SocketMock('''GET / HTTP/1.1
Host: localhost
User-Agent: nengo_gui
''')

        class HandlerClass(server.HttpWsRequestHandler):
            def __init__(self, request, client_address, srv):
                self.method_called = False
                server.HttpWsRequestHandler.__init__(
                    self, request, client_address, srv)

            def http_GET(self):
                self.method_called = True
                assert self.path == '/'
                assert self.headers.get('host').lower() == 'localhost'
                assert self.headers.get('user-agent').lower() == 'nengo_gui'

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler.method_called

    def test_upgrade_to_websocket(self):
        request = SocketMock('''GET /resource_name HTTP/1.1
Upgrade: websocket
Connection: Upgrade
Origin: http://localhost:80
Host: localhost:80
Sec-WebSocket-Key: AQIDBAUGBwgJCgsMDQ4PEC==
Sec-WebSocket-Version: 13
''')

        class HandlerClass(server.HttpWsRequestHandler):
            def ws_default(self):
                pass

            def get_expected_origins(self):
                return ['localhost:80']

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler
        # Ignore errors in decoding because binary WebSocket data might follow
        # response.
        response = request.file_io.written_data.decode(
            'utf-8', errors='ignore')

        assert re.match(r'^HTTP\/\d+\.\d+\s+101.*$', response, re.M)
        assert re.search(r'^Upgrade:\s+.*websocket.*$', response, re.M | re.I)
        assert re.search(r'^Connection:\s+.*upgrade.*$', response, re.M | re.I)
        assert re.search(
            r'^Sec-WebSocket-Accept:\s+OfS0wDaT5NoxF2gqm7Zj2YtetzM=\s+$',
            response, re.M)

    def test_bad_upgrade_to_websocket(self):
        request = SocketMock('''GET /resource_name HTTP/1.1
Upgrade: websocket
Connection: Upgrade
Origin: http://localhost:80
Host: localhost:80
Sec-WebSocket-Key: null
Sec-WebSocket-Version: 13
''')

        class HandlerClass(server.HttpWsRequestHandler):
            def get_expected_origins(self):
                return ['localhost:80']

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler
        response = request.file_io.written_data.decode('utf-8')

        assert re.match(r'^HTTP\/\d+\.\d+\s+400.*$', response, re.M)

    def test_parsing_resource_get(self):
        request = SocketMock('GET /res/file?p1=1&p2=2&p1=0 HTTP/1.1\r\n')

        class HandlerClass(server.HttpWsRequestHandler):
            def http_default(self):
                pass

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler.resource == '/res/file'
        assert handler.query == {'p1': ['1', '0'], 'p2': ['2']}
        assert handler.db == {'p1': '1', 'p2': '2'}

    def test_parsing_resource_post(self):
        content = 'p2=3&p3=0'
        request = SocketMock('''POST /res/file?p1=1&p2=2&p1=0 HTTP/1.1
Content-Type:application/x-www-form-urlencoded; charset=UTF-8
Content-Length: {len}

{content}
'''.format(len=len(content), content=content))

        class HandlerClass(server.HttpWsRequestHandler):
            def http_default(self):
                pass

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler.resource == '/res/file'
        assert handler.query == {'p1': ['1', '0'], 'p2': ['2']}
        assert handler.db == {'p1': '1', 'p2': '3', 'p3': '0'}

    def test_http_dispatch(self):
        request = SocketMock('GET /file HTTP/1.1\r\n')

        class HandlerClass(server.HttpWsRequestHandler):
            called = False
            http_commands = {'/file': 'cmd'}

            def cmd(self):
                self.called = True
                return b''

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler.called

    def test_ws_dispatch(self):
        request = SocketMock('''GET /ws HTTP/1.1
Upgrade: websocket
Connection: Upgrade
Origin: http://localhost:80
Host: localhost:80
Sec-WebSocket-Key: AQIDBAUGBwgJCgsMDQ4PEC==
Sec-WebSocket-Version: 13
''')

        class HandlerClass(server.HttpWsRequestHandler):
            called = False
            ws_commands = {'/ws': 'cmd'}
            server = ServerMock()

            def cmd(self):
                self.called = True

            def get_expected_origins(self):
                return ['localhost:80']

        handler = HandlerClass(request, 'localhost', ServerMock())
        assert handler.called
        assert handler.ws.state in [
            server.WebSocket.ST_CLOSING, server.WebSocket.ST_CLOSED]



# Tutorial 15: The Lorenz Chaotic Attractor

# Differential equations can also give chaotic behaviour.  The classic example
# of this is the Lorenz "butterfly" attractor.  The equations for it are
#
# dx0/dt = sigma * (x1 - x0)
# dx1/dt = - x0 * x2 - x1
# dx2/dt = x0 * x1 - beta * (x2 + rho) - rho
#
# Note: this is a slight transformation from the standard formulation so
#  as to centre the value around the origin. For further information, see
#  http://compneuro.uwaterloo.ca/publications/eliasmith2005b.html
#  "Chris Eliasmith. A unified approach to building and controlling 
#   spiking attractor networks. Neural computation, 7(6):1276-1314, 2005."

# Since there are three dimensions, we can show three different XY plots
# combining the different values in different ways.

import nengo

model = nengo.Network(seed=3)
with model:
    
    x = nengo.Ensemble(n_neurons=600, dimensions=3, radius=30)

    synapse = 0.1
    def lorenz(x):
        sigma = 10
        beta = 8.0/3
        rho = 28
        
        dx0 = -sigma * x[0] + sigma * x[1]
        dx1 = -x[0] * x[2] - x[1]
        dx2 = x[0] * x[1] - beta * (x[2] + rho) - rho
    
        return [dx0 * synapse + x[0],
                dx1 * synapse + x[1],
                dx2 * synapse + x[2]]

    nengo.Connection(x, x, synapse=synapse, function=lorenz)
    



# A Single Neuron Model

# All models in nengo are built inside "networks". Inside a network, you 
# can put more networks, and you can connect networks to each other. You can 
# also put other objects such as neural populations or "ensembles" (which 
# have individual neurons inside of them) inside the networks. For this model, 
# you will make a network with one neuron.

# This model has parameters as described in the book, ensuring that it is an 
# "on" neuron. The neuron will be slightly different each time you run this 
# script, as many parameters are randomly chosen.

# Press the play button to run the simulation.
# The graph on the top left shows the input signal and the graph on the top 
# right shows the value represented by the neuron. The filtered spikes from the 
# neuron are shown in the graph on the bottom right.


# Setup the environment
import numpy as np
import nengo

model = nengo.Network(label='A Single Neuron')
with model:
    #Input
    cos = nengo.Node(lambda t: np.cos(16 * t))
    
    #Ensemble with one neuron
    neuron = nengo.Ensemble(1, dimensions=1,  # Represent a scalar
                encoders=[[1]])   # Sets the neurons firing rate to increase 
                                    # for positive input
    
    #Connecting input to ensemble
    nengo.Connection(cos, neuron)



"""HTTP and WebSocket server implementation."""

import base64
import errno
import hashlib
import logging
import json
import select
import socket
import struct
try:
    import ssl
except ImportError:  # for Python without ssl support
    from . import nossl as ssl
import sys
import threading
import traceback

try:
    from http import server
    from http.cookies import SimpleCookie
    import socketserver
    from urllib.parse import parse_qs, urlparse
except ImportError:  # Python 2.7
    import BaseHTTPServer as server
    from Cookie import SimpleCookie
    import SocketServer as socketserver
    from urlparse import parse_qs, urlparse


logger = logging.getLogger(__name__)
WS_MAGIC = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'


if socket.has_ipv6:
    if hasattr(socket, 'IPPROTO_IPV6'):
        IPPROTO_IPV6 = socket.IPPROTO_IPV6
    elif sys.platform.startswith('win'):
        # See <https://bugs.python.org/issue29515>.
        IPPROTO_IPV6 = 41
    else:
        raise RuntimeError(
            "System does not define IPPROTO_IPV6 despite IPv6 support.")
else:
    IPPROTO_IPV6 = None


class SocketClosedError(IOError):
    pass


class HttpError(Exception):
    def __init__(self, code, msg, headers=(), data=None):
        super(HttpError, self).__init__(msg)
        self.code = code
        self.msg = msg
        self.headers = headers
        if data is None:
            data = b'<h1>' + bytes(self.code) + b'</h1><p>' + msg.encode(
                'utf-8') + b'</p>'
        self.data = data

    def to_response(self):
        return HtmlResponse(self.data, code=self.code, headers=self.headers)


class BadRequest(HttpError):
    def __init__(self):
        super(BadRequest, self).__init__(400, 'Bad request')


class Forbidden(HttpError):
    def __init__(self):
        super(Forbidden, self).__init__(403, 'Forbidden')


class InvalidResource(HttpError):
    def __init__(self, path):
        super(InvalidResource, self).__init__(404, 'Invalid resource: ' + path)


class UpgradeRequired(HttpError):
    def __init__(self, headers):
        super(UpgradeRequired, self).__init__(426, 'Upgrade required', headers)


class InternalServerError(HttpError):
    def __init__(self, msg):
        super(InternalServerError, self).__init__(
            500, 'Internal server error', data=msg.encode('utf-8'))


class HttpResponse(object):
    def __init__(self, data, mimetype='text/html', code=200, headers=()):
        self.data = data
        self.mimetype = mimetype
        self.code = code
        self.headers = headers

    def send(self, request):
        request.send_response(self.code)
        request.send_header('Content-Type', self.mimetype)
        request.send_header('Content-Length', len(self.data))
        if hasattr(request, 'flush_headers'):
            request.flush_headers()
        cookie = request.cookie.output().encode('utf-8')
        if len(cookie) > 0:
            request.wfile.write(cookie)
            request.wfile.write(b'\r\n')
        for header in self.headers:
            request.send_header(*header)
        request.end_headers()
        request.wfile.write(self.data)


class HttpRedirect(HttpResponse):
    def __init__(
            self, location, data=b'', mimetype='text/html', code=303,
            headers=()):
        super(HttpRedirect, self).__init__(
            data=data, mimetype=mimetype, code=code,
            headers=headers + (('Location', location),))
        self.location = location


class HtmlResponse(HttpResponse):
    def __init__(self, body, code=200, headers=()):
        data = b'<html><body>' + body + b'</body></html>'
        super(HtmlResponse, self).__init__(data, code=code, headers=headers)


class JsonResponse(HttpResponse):
    def __init__(self, data, code=200, headers=()):
        super(JsonResponse, self).__init__(
            json.dumps(data).encode('utf-8'), code=code, headers=headers,
            mimetype='application/json')


class DualStackHttpServer(object):
    class Binding(object):
        request_queue_size = 5

        def __init__(self, address_family, address):
            self.address_family = address_family
            self.address = address
            self.socket = socket.socket(
                self.address_family, socket.SOCK_STREAM)
            if self.address_family is socket.AF_INET6:
                self.socket.setsockopt(
                    IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)

        def bind(self):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            self.socket.bind(self.address)
            self.address = self.socket.getsockname()

            host, port = self.address[:2]
            self.name = socket.getfqdn(host)
            self.port = port

        def activate(self):
            self.socket.listen(self.request_queue_size)

        @property
        def host(self):
            return self.address[0]

        @host.setter
        def host(self, value):
            self.address = (value, self.port)

        @property
        def port(self):
            return self.address[1]

        @port.setter
        def port(self, value):
            self.address = (self.host, value)

    def __init__(self, server_address, RequestHandlerClass):
        self.server_name, self.server_port = server_address

        self.bindings = []
        if self.server_name == '':
            self.bindings.append(self.Binding(
                socket.AF_INET, ('', self.server_port)))
            if socket.has_ipv6:
                self.bindings.append(self.Binding(
                    socket.AF_INET6, ('::', self.server_port)))
        else:
            for fam, socktype, _, _, addr in socket.getaddrinfo(
                    *server_address):
                if (fam in (socket.AF_INET, socket.AF_INET6) and
                        socktype in (0, socket.SOCK_STREAM)):
                    self.bindings.append(self.Binding(fam, addr))

        self.RequestHandlerClass = RequestHandlerClass

        self.__is_shut_down = threading.Event()
        self.__shutdown_request = False

        self.server_bind()
        self.server_activate()

    def server_bind(self):
        for b in self.bindings:
            if b.port == 0 and self.server_port != 0:
                # Use same port for all automatically chosen ports
                b.port = self.server_port
            b.bind()
            if self.server_port == 0:
                self.server_port = b.port

    def server_activate(self):
        for b in self.bindings:
            b.activate()

    def serve_forever(self, poll_interval=0.5):
        self.__is_shut_down.clear()
        try:
            while not self.__shutdown_request:
                sockets = [b.socket for b in self.bindings]
                r, _, _ = select.select(sockets, [], [], poll_interval)
                for s in r:
                    self._handle_request_noblock(s)
        finally:
            self.__shutdown_request = False
            self.__is_shut_down.set()

    def shutdown(self):
        self.__shutdown_request = True
        self.__is_shut_down.set()

    def server_close(self):
        for s in self.sockets:
            s.close()

    def handle_request(self):
        raise NotImplementedError()

    def _handle_request_noblock(self, socket):
        request, client_address = socket.accept()
        if self.verify_request(request, client_address):
            try:
                self.process_request(request, client_address)
            except:
                self.handle_error(request, client_address)
                self.shutdown_request(request)
        else:
            self.shutdown_request(request)

    def verify_request(self, request, client_address):
        return True

    def process_request(self, request, client_address):
        self.finish_request(request, client_address)
        self.shutdown_request(request)

    def shutdown_request(self, request):
        try:
            request.shutdown(socket.SHUT_WR)
        except (OSError, socket.error):
            pass
        self.close_request(request)

    def close_request(self, request):
        pass

    def finish_request(self, request, client_address):
        self.RequestHandlerClass(request, client_address, self)

    def handle_error(self, request, client_address):
        logger.exception("Exception while handling request.")


class ManagedThreadHttpServer(
        socketserver.ThreadingMixIn, DualStackHttpServer):
    """Threaded HTTP and WebSocket server that keeps track of its connections
    to allow a proper shutdown."""

    daemon_threads = True  # this ensures all spawned threads exit

    def __init__(self, *args, **kwargs):
        DualStackHttpServer.__init__(self, *args, **kwargs)

        # keep track of open threads, so we can close them when we exit
        self._requests = []
        self._websockets = []

        self._shutting_down = False

    @property
    def requests(self):
        return self._requests[:]

    @property
    def websockets(self):
        return self._websockets[:]

    def create_websocket(self, socket):
        ws = WebSocket(socket)
        self._websockets.append(ws)
        return ws

    def process_request_thread(self, request, client_address):
        thread = threading.current_thread()
        self._requests.append((thread, request))
        socketserver.ThreadingMixIn.process_request_thread(
            self, request, client_address)
        self._requests.remove((thread, request))

    def handle_error(self, request, client_address):
        exc_type, exc_value, _ = sys.exc_info()
        if (exc_type is socket.error and
                exc_value.args[0] in
                [errno.EPIPE, errno.EBADF, errno.ECONNRESET]):
            return  # Probably caused by a server shutdown
        else:
            logger.exception("Server error.")
            DualStackHttpServer.handle_error(self, request, client_address)

    def shutdown(self):
        if self._shutting_down:
            return
        self._shutting_down = True

        for ws in self.websockets:
            ws.close()
        for _, request in self.requests:
            self.shutdown_request(request)

        DualStackHttpServer.shutdown(self)

    def wait_for_shutdown(self, timeout=None):
        """Wait for all request threads to finish.

        Parameters
        ----------
        timeout : float, optional
            Maximum time in seconds to wait for each thread to finish.
        """
        for thread, _ in self.requests:
            if thread.is_alive():
                thread.join(timeout)


class HttpWsRequestHandler(server.BaseHTTPRequestHandler):
    """Base class for request handler that can handle normal and websocket
    requests.

    `http_commands` and `ws_commands` are dictionaries mapping resource names
    (with leading '/') to function names (as string) in this class. These
    functions do not take any arguments except for `self`. All required data
    is defined as attributes on the instance. In addition to the attributes
    provided by the `BaseHTTPRequestHandler` the resource name is provided as
    `resource`, the parsed query string (as dictionary) as `query` and combined
    query string and post fields as `db`. In a websocket command handler
    function `ws` provides access to the websocket.

    If no handler function for a resource was defined, the `http_default` and
    `ws_default` functions will be used.
    """
    http_commands = {}
    ws_commands = {}

    def __init__(self, *args, **kwargs):
        self.resource = None
        self.query = None
        self.db = {}
        self.cookie = SimpleCookie()
        self.ws = None
        server.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def do_POST(self):
        data = self.rfile.read(
            int(self.headers['Content-Length'])).decode('ascii')

        if 'multipart/form-data' in self.headers['Content-Type']:
            raise NotImplementedError()  # TODO
        else:
            self.db = {k: v[0] for k, v in parse_qs(data).items()}

        self.do_GET()

    def do_GET(self):
        parsed = urlparse(self.path)
        self.resource = parsed.path
        self.query = parse_qs(parsed.query)
        self.db.update(
            {k: v[0] for k, v in self.query.items() if k not in self.db})
        if 'Cookie' in self.headers:
            self.cookie.load(self.headers['Cookie'])

        try:
            connection = self.headers.get('Connection', 'close').lower()
            if 'upgrade' in connection:
                self.handle_upgrade()
            else:
                self.http_GET()
        except HttpError as err:
            logger.warning(
                'Error response (%i): %s', err.code, err.msg, exc_info=True)
            err.to_response().send(self)
        except Exception as err:
            logger.exception('Error response')
            err = InternalServerError(
                '<pre>' + traceback.format_exc() + '</pre>')
            err.to_response().send(self)

    def http_GET(self):
        command = self._get_command(self.http_commands, self.resource)
        if command is None:
            response = self.http_default()
        else:
            response = getattr(self, command)()
        response.send(self)

    def http_default(self):
        raise InvalidResource(self.path)

    def handle_upgrade(self):
        upgrade = self.headers.get('Upgrade').lower()
        if upgrade == 'websocket':
            self.upgrade_to_ws()
        else:
            raise BadRequest()

    def get_expected_origins(self):
        raise NotImplementedError()

    def upgrade_to_ws(self):
        response = '''HTTP/1.1 101 Switching Protocols\r\n\
Upgrade: websocket\r\n\
Connection: Upgrade\r\n\
Sec-WebSocket-Accept: {sec}\r\n\
\r\n\
'''
        valid_srv_addrs = self.get_expected_origins()

        try:
            origin = urlparse(self.headers['Origin'])
            assert origin.netloc.lower() in valid_srv_addrs
        except KeyError:
            raise Forbidden()
        except AssertionError:
            raise Forbidden()

        try:
            host = self.headers['Host'].lower()
            assert host in valid_srv_addrs
            key = self.headers['Sec-WebSocket-Key']
            assert len(base64.b64decode(key)) == 16
        except KeyError:
            raise BadRequest()
        except AssertionError:
            raise BadRequest()

        if self.headers['Sec-WebSocket-Version'] != '13':
            raise UpgradeRequired(['Sec-WebSocket-Version: 13'])

        sec = base64.b64encode(hashlib.sha1(
            (key + WS_MAGIC).encode('ascii')).digest()).decode('ascii')
        _sendall(self.request, response.format(sec=sec).encode('utf-8'))

        self.ws = self.server.create_websocket(self.request)
        self.ws.set_blocking(False)

        command = self._get_command(self.ws_commands, self.resource)
        if command is None:
            self.ws_default()
        else:
            getattr(self, command)()
        self.ws.close()

    def ws_default(self):
        raise InvalidResource(self.path)

    @classmethod
    def _get_command(cls, commands, path):
        if not path.startswith('/'):
            path = '/' + path

        while len(path) > 0:
            if path in commands:
                return commands[path]
            path = path.rsplit('/', 1)[0]
        if '/' in commands:
            return commands['/']
        return None


class WebSocket(object):
    ST_OPEN, ST_CLOSING, ST_CLOSED = range(3)

    def __init__(self, socket):
        self.socket = socket
        self._buf = bytearray([])
        self.state = self.ST_OPEN

    def set_timeout(self, timeout):
        self.socket.settimeout(timeout)

    def set_blocking(self, flag):
        self.socket.setblocking(flag)

    def _read(self):
        try:
            self._buf = self._buf + bytearray(self.socket.recv(512))
        except ssl.SSLError as e:
            if e.errno == 2:
                # Corresponds to SSLWantReadError which only exists in Python
                # 2.7.9+ and 3.3+.
                pass
            else:
                raise
        except socket.error as e:
            if e.errno in [errno.EDEADLK, errno.EAGAIN, 10035]:
                # no data available
                pass
            elif e.errno == errno.EBADF:
                raise SocketClosedError("Cannot read from closed socket.")
            else:
                raise

    def read_frame(self):
        try:
            self._read()
            frame, size = WebSocketFrame.parse(self._buf)
            self._buf = self._buf[size:]
            if not self._handle_frame(frame):
                return frame
        except ValueError:
            return None
        except socket.timeout:
            return None

    def _handle_frame(self, frame):
        if frame.opcode == WebSocketFrame.OP_CLOSE:
            if self.state not in [self.ST_CLOSING, self.ST_CLOSED]:
                self.close()
            raise SocketClosedError("Websocket has been closed")
        elif frame.opcode == WebSocketFrame.OP_PING:
            if self.state == self.ST_OPEN:
                pong = WebSocketFrame(
                    fin=1, rsv=0, opcode=WebSocketFrame.OP_PONG, mask=0,
                    data=frame.data)
                _sendall(self.socket, pong.pack())
            return True
        elif frame.opcode == WebSocketFrame.OP_PONG:
            return True
        else:
            return False

    def close(self):
        if self.state not in [self.ST_CLOSING, self.ST_CLOSED]:
            self.state = self.ST_CLOSING
            close_frame = WebSocketFrame(
                fin=1, rsv=0, opcode=WebSocketFrame.OP_CLOSE, mask=0, data=b'')
            try:
                _sendall(self.socket, close_frame.pack())
            except socket.error as err:
                if err.errno in [errno.EPIPE, errno.EBADF]:
                    pass

    def write_frame(self, frame):
        if self.state != self.ST_OPEN:
            raise SocketClosedError("Connection not open.")

        try:
            _sendall(self.socket, frame.pack())
        except socket.error as e:
            if e.errno == errno.EPIPE:  # Broken pipe
                raise SocketClosedError("Cannot write to socket.")
            else:
                raise

    def write_text(self, text):
        self.write_frame(WebSocketFrame.create_text_frame(text))

    def write_binary(self, data):
        self.write_frame(WebSocketFrame.create_binary_frame(data))


def _sendall(socket, data):
    bytes_sent = 0
    while bytes_sent < len(data):
        select.select((), (socket,), ())  # ensure socket is not full
        bytes_sent += socket.send(data[bytes_sent:])


class WebSocketFrame(object):
    __slots__ = ['fin', 'rsv', 'opcode', 'mask', 'data']

    OP_CONT = 0x0
    OP_TEXT = 0x1
    OP_BIN = 0x2
    OP_CLOSE = 0x8
    OP_PING = 0x9
    OP_PONG = 0xA

    def __init__(self, fin, rsv, opcode, mask, data):
        self.fin = fin
        self.rsv = rsv
        self.opcode = opcode
        self.mask = mask
        self.data = data

    @classmethod
    def parse(cls, data):
        try:
            offset = 0

            fin = (data[0] >> 7) & 0x1
            rsv = (data[0] >> 4) & 0x07
            opcode = data[0] & 0x0F
            masked = (data[1] >> 7) & 0x01
            datalen = data[1] & 0x7F
            mask = b'\x00\x00\x00\x00'

            offset += 2

            if datalen == 126:
                datalen = cls._to_int(data[offset:offset+2])
                offset += 2
            elif datalen == 127:
                datalen = cls._to_int(data[offset:offset+8])
                offset += 8

            if masked:
                mask = data[offset:offset+4]
                offset += 4

            size = offset + datalen
            masked_data = data[offset:size]
            if len(masked_data) < datalen:
                raise IndexError()
            unmasked_data = [masked_data[i] ^ mask[i % 4]
                             for i in range(len(masked_data))]
            data = bytearray(unmasked_data)
            if opcode == cls.OP_TEXT:
                data = data.decode('ascii')

            return cls(fin, rsv, opcode, mask, data), size
        except IndexError:
            raise ValueError('Frame incomplete.')

    @classmethod
    def _to_int(cls, data):
        value = 0
        for b in data:
            value = (value << 8) + b
        return value

    def pack(self):
        code = (self.fin & 0x01) << 7
        code |= (self.rsv & 0x07) << 4
        code |= self.opcode & 0x0F

        datalen = len(self.data)
        mask_bit = ((self.mask != 0) & 0x01) << 7
        if datalen < 126:
            header = struct.pack('!BB', code, datalen | mask_bit)
        elif datalen <= 0xFFFF:
            header = struct.pack('!BBH', code, 126 | mask_bit, datalen)
        else:
            header = struct.pack('!BBQ', code, 127 | mask_bit, datalen)

        data = self.data

        return header + data

    @classmethod
    def create_text_frame(cls, text, mask=0):
        return cls(1, 0, cls.OP_TEXT, mask, text.encode('utf-8'))

    @classmethod
    def create_binary_frame(cls, data, mask=0):
        return cls(1, 0, cls.OP_BIN, mask, data)



# Tutorial 13: Oscillators

# If we do differential equations in multiple dimensions, we can get
# oscillators.  This gives an Ensemble of neurons that can produce patterns
# of behaviour all on its own without any external input.  For example, here
# is a standard cycle in two dimensions:
#    dx0/dt = -x1 * s + x0 * (r - x0**2 - x1**2)
#    dx1/dt =  x0 * s + x1 * (r - x0**2 - x1**2)
# where r is the radius of the circle and s is the speed (in radians per second).

# As discussed in the previous tutorial, we can convert this into a Nengo
# model.  In this case there is no input connection, so all we have to do
# is multiply by the synapse and add the original value.

# Here we introduce a new kind of plot.  The XY-value plot shows the same
# information as the normal Value plot, but plots the two dimensions together
# rather than using time to be the x-axis.  This can be convenient for
# representing multidimensional data.

# Try adjusting the r value to 0.5.  Try 1.5.  What about 0?
# Try adjusting the speed s.  What happens when it is very slow (0.5)?  0.1?

import nengo

model = nengo.Network()
with model:

    x = nengo.Ensemble(n_neurons=200, dimensions=2)

    synapse = 0.1
    def oscillator(x):
        r = 1
        s = 6
        return [synapse * (-x[1] * s + x[0] * (r - x[0]**2 - x[1]**2)) + x[0],
                synapse * ( x[0] * s + x[1] * (r - x[0]**2 - x[1]**2)) + x[1]]

    nengo.Connection(x, x, synapse=synapse, function=oscillator)





# Tutorial 14: Controlled Oscillator

# Here we do the exact same oscillator as in the previous example, but we
# introduce a new dimension that lets us control the speed of the oscillation

# We use the same differential equation as before:
#    dx0/dt = -x1 * s + x0 * (r - x0**2 - y0**2)
#    dx1/dt =  x0 * s + x1 * (r - x0**2 - y0**2)
# where r is the radius of the circle and s is the speed (in radians per
# second).

# But, in this case, we make the Ensemble be 3-dimensional and use the third
# dimension (x[2]) to represent s.  You can control it with a separate input.
# This shows how neurons can affect the pattern of activity of another
# group of neurons.


import nengo

model = nengo.Network()
with model:

    x = nengo.Ensemble(n_neurons=400, dimensions=3)

    synapse = 0.1

    def oscillator(x):
        r = 1
        s = 10 * x[2]
        return [synapse * -x[1] * s + x[0] * (r - x[0]**2 - x[1]**2) + x[0],
                synapse *  x[0] * s + x[1] * (r - x[0]**2 - x[1]**2) + x[1]]

    nengo.Connection(x, x[:2], synapse=synapse, function=oscillator)

    stim_speed = nengo.Node(0)
    speed = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stim_speed, speed)
    nengo.Connection(speed, x[2])



import nengo
import numpy as np

model = nengo.Network()
with model:
    a = nengo.Ensemble(n_neurons=40, dimensions=1)
    
    def stim_function(t):
        return np.sin(t*2*np.pi)
    stim = nengo.Node(stim_function)
    
    nengo.Connection(stim, a, synapse=None)
    



# # Nengo Example: 2-Dimensional Representation
#
# Ensembles of neurons represent information. In Nengo, we represent that
# information with real-valued vectors -- lists of numbers. In this example,
# we will represent a two-dimensional vector with a single ensemble of leaky
# integrate-and-fire neurons.

import nengo
import numpy as np

model = nengo.Network()
with model:
    neurons = nengo.Ensemble(n_neurons=100, dimensions=2)

    stim = nengo.Node([0, 0])
    nengo.Connection(stim, neurons)



import numpy as np
import nengo

model = nengo.Network()
with model:
    stimulus = nengo.Node(lambda t: (np.sin(t), np.cos(t)))
    ens = nengo.Ensemble(n_neurons=1000, dimensions=2)
    nengo.Connection(stimulus, ens)

if __name__ == '__main__':
    import nengo_gui
    nengo_gui.GUI(__file__).start()



from __future__ import print_function

import os
import time

import pytest
from selenium.webdriver import ActionChains

from nengo_gui import testing_tools as tt

test_files = tt.folder_location('examples/basics')


@pytest.mark.parametrize('test_file', test_files)
def test_basic_functionality(driver, test_file):
    try:
        # Test page response by clicking the reset button and applying
        # new code to ace-editor
        tt.reset_page(driver)
        time.sleep(0.5)
        tt.update_editor(driver, test_file)
        tt.mouse_scroll(driver, 500)
        ens_elements = driver.find_elements_by_xpath('//*[@class="ens"]')
        assert (len(ens_elements) > 0)
        side_script = ("var right = document.getElementById('rightpane');\n"
                       "right.style.width = '200px';\n")
        driver.execute_script(side_script)

        # Creates graph objects by right clicking and selecting
        actions = ActionChains(driver)
        elements = ['node', 'ens']
        for elem in elements:
            node = driver.find_element_by_xpath('//*[@class="'+elem+'"]')
            actions = ActionChains(driver)
            actions.move_to_element(node)
            actions.context_click()
            actions.perform()
            time.sleep(1)
            actions = ActionChains(driver)

            menu = driver.find_element_by_xpath(
                '//*[@class="dropdown-menu"]/li[1]')
            actions.move_to_element(menu)
            actions.click()
            actions.perform()
            time.sleep(0.5)

        graph_elements = driver.find_elements_by_xpath('//*[@class="graph"]')

        assert len(graph_elements) > 0
        tt.start_stop_sim(driver)
        time.sleep(1.5)
        tt.start_stop_sim(driver)
        time.sleep(0.5)

        ticker = driver.find_element_by_xpath('//*[@id="ticks_tr"]/td')
        sim_time = ticker.get_attribute('textContent')

        assert (float(sim_time) > 0)

    except:
        # Travis Only: On fail takes screenshot and uploads it to imgur
        if('TRAVIS' in os.environ):
            tt.imgur_screenshot(driver)
        raise



from __future__ import print_function

import os.path
import socket
import threading

import pytest
from selenium import webdriver

import nengo_gui
from nengo_gui import guibackend
from nengo_gui.gui import BaseGUI


@pytest.yield_fixture(scope="session")
def gui():
    host, port = ('localhost', 0)
    server_settings = guibackend.GuiServerSettings((host, port))
    model_context = guibackend.ModelContext(
        filename=os.path.join(nengo_gui.__path__[0], 'examples', 'default.py'))
    gui = BaseGUI(model_context, server_settings)
    server_thread = threading.Thread(target=gui.start)
    server_thread.daemon = True
    server_thread.start()
    port = gui.server.server_port

    started = False
    while server_thread.is_alive() and not started:
        try:
            s = socket.create_connection((host, port), 0.1)
            started = True
        except:
            pass
        else:
            s.close()

    yield gui

    gui.server.shutdown()
    gui.server.wait_for_shutdown(0.05)


@pytest.yield_fixture(scope="session")
def driver(gui):
    driver = webdriver.Firefox()
    driver.get('http://localhost:{port}/'.format(port=gui.server.server_port))
    driver.maximize_window()

    assert driver.title != "Problem loading page"
    yield driver

    driver.quit()



# Nonlinear Transformations

# The real world is filled with non-linearities and so dealing 
# with them often requires nonlinear computation. This model shows 
# how to compute nonlinear functions using Nengo 2.0. The two nonlinear
# functions demonstrated in this model are 'multiplication' and 'squaring'.

# The parameters of the model are as described in the book. The model has five 
# ensembles: two input ensembles (X and Y), a 2D combined ensemble (vector2D), 
# and the output ensembles result_square and result_product which store the 
# square and product of the inputs respectively.

# Two varying scalar values are used for the two input signals that drive 
# activity in ensembles A and B. For multiplication, we project both 
# inputs independently into a 2D space, and then decode a nonlinear 
# transformation of that space (i.e., the product) into an ensemble 
# (result_product). The model also squares the value of the first input 
# (inputX) encoded in an ensemble (X), in the output of another ensemble 
# (result_square).

# The two functions product(x) and square(x) are defined to serve the same 
# purpose as entering the expressions in the "Expression" field in the 
# "User-defined Function" dialog box in Nengo 1.4 as described in the book.

# Press the play button to run the simulation.
# The input signals chosen clearly show that the model works well. 
# The result_product graph shows the product of the inputX & inputY, and 
# the result_square graph shows the squre of inputX. You can see in the graphs 
# that when inputX is zero, both the product and the square are also zero. You 
# can use the sliders to change the input values provided by the inputX and 
# inputY nodes to test the model.

# To inspect any values in the graphs, you can hover your mouse over the graph
# and the x and t values will be displayed near the axis.  This can be easier
# when you pause the simulation.  When you pause, you can also drag the blue
# box at the bottom of the screen to slide backwards and forward through time.

# Setup the environment
import nengo
from nengo.dists import Choice
from nengo.utils.functions import piecewise

model = nengo.Network(label='Nonlinear Function')
with model:
    #Input - Piecewise step functions
    inputX = nengo.Node(piecewise({0: -0.75, 1.25: 0.5, 2.5: -0.75, 3.75: 0}))
    inputY= nengo.Node(piecewise({0: 1, 1.25: 0.25, 2.5: -0.25, 3.75: 0.75}))
    
    #Five ensembles containing LIF neurons
    X = nengo.Ensemble(100, dimensions=1, radius=1)     # Represents inputX
    Y = nengo.Ensemble(100, dimensions=1, radius=1)     # Represents inputY
    # 2D combined ensemble
    vector2D = nengo.Ensemble(224, dimensions=2, radius=2)  
    # Represents the square of X
    result_square = nengo.Ensemble(100, dimensions=1, radius=1) 
    # Represents the product of X and Y
    result_product = nengo.Ensemble(100, dimensions=1, radius=1)  
    
    #Connecting the input nodes to the appropriate ensembles
    nengo.Connection(inputX, X)
    nengo.Connection(inputY, Y)
    
    #Connecting input ensembles A and B to the 2D combined ensemble
    nengo.Connection(X, vector2D[0])
    nengo.Connection(Y, vector2D[1])
    
    #Defining a function that computes the product of two inputs
    def product(x):
        return x[0] * x[1]
    
    #Defining the squaring function
    def square(x):
        return x[0] * x[0]
    
    #Connecting the 2D combined ensemble to the result ensemble 
    nengo.Connection(vector2D, result_product, function=product)
    
    #Connecting ensemble A to the result ensemble
    nengo.Connection(X, result_square, function=square)



# Routed Sequencing with Cleanup all Memory

# The previous model on Routed Sequencing with cleanup memory used a scalar 
# ensemble to project only state 'A'. In this model, you will project all the 
# states onto an ensemble of neurons as the state cycles through a five element 
# sequence.

# This model has parameters as described in the book. It extends the routed 
# sequencing model by creating a scalar ensemble 'cleanup' and then projecting 
# all the states on to this ensemble using a transformation matrix 'pd_new', 
# specified by the vectors in the vocabulary.

# Press the play button to run the simulation. 
# The graph in the middle shows the semantic pointer representation of the 
# values stored in state and the graph on the bottom-right shows the response of
# the cleanup population. and the graph on the top-right shows the utility 
# (similarity) of the current basal ganglia input (i.e., state) with the 
# possible vocabulary vectors.

# Since the cleanup operation is similar to a dot product between the state and 
# the defined vocabulary vectors, the value of the cleanup population in a 
# particular dimension rises only when the value of state (top-right graph) 
# corresponds to that particular dimension.

# Setup the environment
import numpy as np 
import nengo
from nengo import spa                #import spa related packages
from nengo.spa import Vocabulary

#Number of dimensions for the Semantic Pointers
dimensions=16

#Make a model object with the SPA network
model = spa.SPA(label='Routed_Sequence with cleanupAll', seed=7)

with model:
    #Specify the modules to be used
    model.state = spa.State(dimensions=dimensions, feedback=1, 
                                        feedback_synapse=0.01)
    model.vision = spa.State(dimensions=dimensions) 
    #Specify the action mapping
    actions = spa.Actions(
        'dot(vision, START) --> state = vision',
        'dot(state, A) --> state = B',
        'dot(state, B) --> state = C',
        'dot(state, C) --> state = D',
        'dot(state, D) --> state = E',
        'dot(state, E) --> state = A'
    )
    
    #Creating the BG and Thalamus components that confirm to the specified rules
    model.BG = spa.BasalGanglia(actions=actions)
    model.thal = spa.Thalamus(model.BG)
    
    #Changing the seed of this RNG to change the vocabulary
    rng = np.random.RandomState(7)
    vocab = Vocabulary(dimensions=dimensions, rng=rng)
    
    #Creating the transformation matrix (pd_new) and cleanup ensemble (cleanup)  
    vsize = len((model.get_output_vocab('state').keys))

    pd = []
    for item in range(vsize):
        pd.append(model.get_output_vocab('state').keys[item])  
          
    pd_new = []
    for element in range(vsize):
        pd_new.append([vocab[pd[element]].v.tolist()]) 

    model.cleanup = nengo.networks.EnsembleArray(
        n_neurons=60, n_ensembles=vsize)

    #Function that provides the model with an initial input semantic pointer.
    def start(t):
        if t < 0.4:
            return '0.8*START+D'
        else:
            return '0'

    #Input
    model.input = spa.Input(vision=start)
    
    #Projecting the state on to the cleanup ensemble using a transformation 
    #matrix 'pd'.
    for i in range(5):
        nengo.Connection(model.state.output, model.cleanup.input[i],
                                            transform=pd_new[i])  
                                            



"""Password hashing functions replicating bcrypt API."""

from __future__ import print_function

import binascii
from getpass import getpass
import hashlib
import os


def gensalt(size=16):
    return binascii.hexlify(os.urandom(size)).decode('ascii')


def hashpw(password, salt, algorithm='sha1'):
    h = hashlib.new(algorithm)
    h.update(password.encode('utf-8'))
    h.update(salt.encode('ascii'))
    return algorithm + ':' + salt + ':' + h.hexdigest()



def checkpw(password, hashed):
    algorithm, salt, _ = hashed.split(':')
    return hashpw(password, salt, algorithm) == hashed


def prompt_pw():
    while True:
        p0 = getpass("Enter password: ")
        p1 = getpass("Enter password: ")
        if p0 == p1:
            return p0
        print("Passwords do not match. Please try again.")



import time

import pytest

from nengo_gui import testing_tools as tt


@pytest.mark.skipif(True, reason="Halts future tests")
def test_pdb_error(driver):
    test_file = """
import nengo
import pdb

model = nengo.Network()
with model:
    pdb.set_trace()
    stim = nengo.Node([0])
    a = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stim, a)
    """
    tt.reset_page(driver)
    tt.update_editor(driver, test_file)
    time.sleep(0.5)
    tt.start_stop_sim(driver)
    time.sleep(1.5)
    tt.start_stop_sim(driver)
    time.sleep(0.5)

    ticker = driver.find_element_by_xpath('//*[@id="ticks_tr"]/td')
    sim_time = ticker.get_attribute('textContent')

    assert (float(sim_time) > 0)



import numpy as np
import struct
import collections

try:
    from nengo.processes import Process
except ImportError:

    class Process(object):
        pass

from nengo.utils.compat import is_iterable

from nengo_gui.components.component import Component


class OverriddenOutput(Process):
    def __init__(self, base_output, to_client=None, from_client=None):
        super(OverriddenOutput, self).__init__()
        self.base_output = base_output
        self.to_client = to_client
        self.from_client = from_client

    def make_step(self, shape_in, shape_out, dt, rng):
        size_out = shape_out[0] if is_iterable(shape_out) else shape_out

        if self.base_output is None:
            f = self.passthrough
        elif isinstance(self.base_output, Process):
            f = self.base_output.make_step(shape_in, shape_out, dt, rng)
        else:
            f = self.base_output
        return self.Step(size_out, f,
                         to_client=self.to_client,
                         from_client=self.from_client)

    @staticmethod
    def passthrough(t, x):
        return x

    class Step(object):
        def __init__(self, size_out, f, to_client=None, from_client=None):
            self.size_out = size_out
            self.f = f
            self.to_client = to_client
            self.from_client = from_client

            self.last_time = None
            self.struct = struct.Struct('<%df' % (1 + self.size_out))
            self.value = np.zeros(size_out, dtype=np.float64)

        def __call__(self, t, *args):
            # Stop overriding if we've reset
            if self.last_time is None or t < self.last_time:
                self.from_client[:] = np.nan
            self.last_time = t

            val_idx = np.isnan(self.from_client)

            if callable(self.f):
                self.value[:] = np.atleast_1d(self.f(t, *args))
                if self.to_client is not None:
                    self.to_client.append(self.struct.pack(t, *self.value))
            else:
                self.value[:] = self.f

            # Override values from the client
            self.value[~val_idx] = self.from_client[~val_idx]
            return self.value


class Slider(Component):
    """Input control component. Exclusively associated to Nodes"""

    config_defaults = dict(max_value=1, min_value=-1,
                           **Component.config_defaults)

    def __init__(self, node):
        super(Slider, self).__init__()
        self.node = node
        self.base_output = node.output

        self.to_client = collections.deque()
        self.from_client = np.zeros(node.size_out, dtype=np.float64) * np.nan
        self.override_output = OverriddenOutput(
            self.base_output, self.to_client, self.from_client)
        self.start_value = np.zeros(node.size_out, dtype=np.float64)
        if not (self.base_output is None or
                callable(self.base_output) or
                isinstance(self.base_output, Process)):
            self.start_value[:] = self.base_output

    def attach(self, page, config, uid):
        super(Slider, self).attach(page, config, uid)
        self.label = page.get_label(self.node)

    def add_nengo_objects(self, page):
        if Process.__module__ == "nengo_gui.components.slider":
            self.node.output = self.override_output.make_step(
                shape_in=None, shape_out=self.node.size_out, dt=None, rng=None)
        elif page.settings.backend == 'nengo_spinnaker':
            # TODO: this should happen for any backend that does not support
            #  Processes
            self.node.output = self.override_output.make_step(
                shape_in=None, shape_out=self.node.size_out, dt=None, rng=None)
        else:
            self.node.output = self.override_output

    def remove_nengo_objects(self, page):
        self.node.output = self.base_output

    def javascript(self):
        info = dict(uid=id(self), n_sliders=self.node.size_out,
                    label=self.label,
                    start_value=[float(x) for x in self.start_value])
        json = self.javascript_config(info)
        return 'new Nengo.Slider(main, sim, %s);' % json

    def update_client(self, client):
        while len(self.to_client) > 0:
            to_client = self.to_client.popleft()
            client.write_binary(to_client)

    def message(self, msg):
        index, value = msg.split(',')
        index = int(index)
        if value == 'reset':
            self.from_client[index] = np.nan
        else:
            self.from_client[index] = float(value)

    def code_python_args(self, uids):
        return [uids[self.node]]



# Nengo Example: Two Neurons
#
# This demo shows how to construct and manipulate a complementary pair of
# neurons.
#
# These are leaky integrate-and-fire (LIF) neurons. The neuron tuning
# properties have been selected so there is one on and one off neuron.
#
# One neuron will increase for positive input, and the other will decrease.
# This can be thought of as the simplest population that is able to give a
# reasonable representation of a scalar value.

import nengo
import numpy as np

model = nengo.Network()
with model:
    neurons = nengo.Ensemble(n_neurons=2, dimensions=1)
    # Set the intercepts at .5
    neurons.intercepts=nengo.dists.Uniform(-.5, -.5)
    # Set the max firing rate at 100hz
    neurons.max_rates=nengo.dists.Uniform(100, 100)
    # One 'on' and one 'off' neuron
    neurons.encoders=[[1],[-1]]
    
    # make a
    stim = nengo.Node([0])

    nengo.Connection(stim, neurons, synapse=0.01)



import time

from selenium.webdriver import ActionChains

from nengo_gui import testing_tools as tt


def test_data_to_csv(driver):
    tt.reset_page(driver)

    tt.update_editor(driver, """
import nengo
model = nengo.Network()
with model:
    stim = nengo.Node([0])
    a = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stim, a)
    """)

    time.sleep(1)

    actions = ActionChains(driver)
    elements = ['node', 'ens']
    for elem in elements:
        for x in range(2):
            node = driver.find_element_by_xpath('//*[@class="'+elem+'"]')

            actions = ActionChains(driver)
            actions.move_to_element(node)
            actions.context_click()
            actions.perform()
            time.sleep(1)
            actions = ActionChains(driver)

            menu = driver.find_element_by_xpath(
                '//*[@class="dropdown-menu"]/li[%d]' % (x+1))
            actions.move_to_element(menu)
            actions.click()
            actions.perform()

            time.sleep(0.5)
    result = driver.execute_script("""
a = function() {
    var data_set = Nengo.Component.components;
    data_set.forEach(function(data, index) {
        if (data.constructor === Nengo.Value) {
            data.data_store.reset();
            data.data_store.push([1, 10+index]);
        }
    });
    return data_to_csv(data_set);
};
return a();
    """)

    test_data = ("Graph Name,stim,a\n"
                 "Times,Dimension1,Dimension1\n"
                 "1,11,12")

    assert result == test_data



# Nengo Example: Controlled Oscillator
#
# The controlled oscillator is an oscillator with an extra input that controls
# the frequency of the oscillation.

import nengo

tau = 0.1   # Post-synaptic time constant for feedback
w_max = 10  # Maximum frequency is w_max/(2*pi)

model = nengo.Network()
with model:
    # The ensemble for the oscillator
    oscillator = nengo.Ensemble(500, dimensions=3, radius=1.7)

    # The feedback connection
    def feedback(x):
        x0, x1, w = x  # These are the three variables stored in the ensemble
        return x0 + w*w_max*tau*x1, x1 - w*w_max*tau*x0, 0
    nengo.Connection(oscillator, oscillator, function=feedback, synapse=tau)

    # The ensemble for controlling the speed of oscillation
    frequency = nengo.Ensemble(100, dimensions=1)

    nengo.Connection(frequency, oscillator[2])

    # We need a quick input at the beginning to start the oscillator
    initial_stim = nengo.Node(lambda t: 1 if t < 0.15 else 0)
    nengo.Connection(initial_stim, oscillator[0])

    # Vary the speed over time
    stim_frequency = nengo.Node([1])

    nengo.Connection(stim_frequency, frequency)



# Structured Representations using SPA

# Now we will build this model again, using the spa (semantic pointer 
# architecture) package built into Nengo 2.0. You will see that using the spa 
# package considerably simplifies model construction and visualization through 
# nengo_gui.

# Press the play button to run the simulation.
# The graphs A and B show the semantic pointer representations in objects A and 
# B respectively. Graphs labelled C show the result of the convolution operation 
# (left - shows the semantic pointer representation in object C, right - shows 
# the similarity with the vectors in the vocabulary). The graphs labelled Sum 
# show the sum of A and B as represented by the object Sum (left - shows the 
# semantic pointer representation in Sum, right - shows high similarity with 
# vecotors A and B).

# Setup the environment
import nengo
import nengo.spa as spa
from nengo.spa import Vocabulary
import numpy as np

D = 32  # the dimensionality of the vectors

#Creating a vocabulary
rng = np.random.RandomState(0)
vocab = Vocabulary(dimensions=D, rng=rng)
vocab.add('C', vocab.parse('A * B'))

model = spa.SPA(label="structure", vocabs=[vocab])
with model:
    model.A = spa.State(D)
    model.B = spa.State(D)
    model.C = spa.State(D, feedback=1)
    model.Sum = spa.State(D)

    actions = spa.Actions(
        'C = A * B',
        'Sum = A',
        'Sum = B'
        )

    model.cortical = spa.Cortical(actions)
    
    #Input
    model.input = spa.Input(A='A', B='B')    



import time

from selenium.webdriver import ActionChains

from nengo_gui import testing_tools as tt


def test_config_persists(driver):
    tt.reset_page(driver)
    menu = driver.find_element_by_id('Config_menu')
    actions = ActionChains(driver)
    actions.move_to_element(menu)
    actions.click()
    actions.perform()
    time.sleep(0.5)
    button = driver.find_element_by_id("Config_button")
    actions = ActionChains(driver)
    actions.move_to_element(button)
    actions.click()
    actions.perform()
    time.sleep(1.0)

    actions = ActionChains(driver)
    checkbox = driver.find_element_by_id("zoom-fonts")
    actions.move_to_element(checkbox)
    actions.click()
    actions.perform()
    time.sleep(0.5)

    val = driver.execute_script("return localStorage.getItem('ng.zoom_fonts')")
    assert val == 'true'

    actions.perform()  # do it again to go back to default (False)
    val = driver.execute_script("return localStorage.getItem('ng.zoom_fonts')")
    assert val == 'false'



# 2D Decision Integrator

# This is a model of perceptual decision making using a two dimensional 
# integrator. As mentioned in the book, the goal is to construct a simple model 
# of perceptual decision making without being concerned with establishing a 
# detailed connection to neurobiology (we have done this elsewhere).

# Rather than having two different integrators for each dimension, you will 
# build the model using a single two dimensional integrator. This integrator 
# can be used irrespective of the task demands since it effectively integrates 
# in every direction simultaneously. This is neurally more efficient due to the 
# reasons explained in the book.

# The model has four ensembles: MT representing the motion area, LIP 
# representing the lateral intraparietal area, input and output of the 2D 
# integrator. The parameters used in the model are as described in the book. 
# The 2D integrator resides in LIP. As discussed in the book an integrator 
# requires two connections: here, the input from MT to LIP and the feedback 
# connection from LIP to LIP.

# Here, you will provide a an input of (-0.5, 0.5) to the model spanning over a 
# period of 6 seconds to observe the model behaviour. In order to inject noise 
# while the simulation runs, you can use the 'noise' parameter when creating 
# ensembles as shown. The reason for injecting noise is explained in the book.

# Press the play button to run the simulation.
# The output plot on the bottom-right shows the output of the 2D decision 
# integrator which is represented by a single two dimensional output ensemble. 
# You can see that as MT encodes the input over time, LIP slowly moves towards 
# the same direction as it acuumulates evidence that there is sustained motion 
# in that direction.

# Thus MT moves LIP in the right direction and once past a certain threshold, 
# the output neurons start firing. To visualize this: 
# 1) Select 'spikes' from the right-click menu of the output ensemble. This will
#    display a spike plot. 
# 2) Run the simulation and then slide the blue box in the simulation control 
#    bar backwards. 
# 3) You will see that the spikes become stronger once past a certain threshold 
#    (i.e., when LIP starts following MT)

#Setup the environment
import numpy as np
import nengo
from nengo.processes import WhiteNoise
from nengo.utils.functions import piecewise
from nengo.utils.matplotlib import rasterplot
from nengo.dists import Uniform

model = nengo.Network(label='2D Decision Integrator', seed=11)
with model:
    #Input
    input1 = nengo.Node(-0.5)
    input2 = nengo.Node(0.5)
    
    #Ensembles 
    input = nengo.Ensemble(100, dimensions=2)
    MT = nengo.Ensemble(100, dimensions=2, 
                                    noise=WhiteNoise(dist=Uniform(-0.3,0.3)))
    LIP = nengo.Ensemble(200, dimensions=2, 
                                    noise=WhiteNoise(dist=Uniform(-0.3,0.3)))
    output = nengo.Ensemble(100, dimensions=2, 
                                    noise=WhiteNoise(dist=Uniform(-0.3,0.3)))
     
    weight = 0.1
    #Connecting the input signal to the input ensemble
    nengo.Connection(input1, input[0], synapse=0.01)   
    nengo.Connection(input2, input[1], synapse=0.01) 
    
    #Providing input to MT ensemble
    nengo.Connection(input, MT, synapse=0.01) 
    
    #Connecting MT ensemble to LIP ensemble
    nengo.Connection(MT, LIP, transform=weight, synapse=0.1) 
    
    #Connecting LIP ensemble to itself
    nengo.Connection(LIP, LIP, synapse=0.1) 
    
    #Connecting LIP population to output
    nengo.Connection(LIP, output, synapse=0.01) 


from __future__ import print_function
import os
import inspect
import nengo_gui
import time


def update_editor(driver, nengoCode):

    """Inserts a string which represents code into ace editor

    Example:
    driver = webdriver.firefox()
    driver.get(localhost:8080/)
    code_string = "print hello world"
    update_editor(driver, code_string)
    Ace editor will update with "print hello world"
    """

    nengoCode = nengoCode.replace("\n", "\\n").replace("\r", "\\r")
    js = "var editor = ace.edit('editor');editor.setValue('"+nengoCode+"');"
    driver.execute_script(js)
    time.sleep(1)


def reset_page(driver):

    """Resets the Nengo gui page

    Example:
    driver = webdriver.firefox()
    driver.get(localhost:8080/)
    reset_page(driver)
    The page then resets
    """
    driver.execute_script("toolbar.reset_model_layout();");
    time.sleep(0.3);


def start_stop_sim(driver):
    """Clicks the start simulation start button"""
    play_button = driver.find_element_by_xpath('//*[@id="pause_button"]')
    play_button.click()


def folder_location(var_path, indiv_file=None):

    """Returns a list of the raw text from a python file in var_path

     Example:
     File Structure:
     --nengo_gui/
       --nengo_gui/
              --stuff/
                   --tests/
                       bar.py
                       foo.py

     foo.py:
       print "hello world"
     bar.py:
       print "goodbye world"

     Code:
     files_raw = folder_location('stuff/tests')
     Returns list of ['print "hello world"','print "goodbye world"']
    """

    folder = os.path.dirname(inspect.getfile(nengo_gui))

    test_folder = os.path.join(folder, var_path)

    test_files = os.listdir(test_folder)
    test_files = filter((lambda x: ((x.count('.') == 1) and ('.py' in x))),
                        test_files)
    if(indiv_file is not None):
        test_files = filter((lambda x: (x == indiv_file)), test_files)
    test_files = map((lambda file_: os.path.join(test_folder, file_)),
                     test_files)
    test_files = map((lambda file_: open(file_, 'r').read()), test_files)
    test_files = map((lambda raw_file: raw_file.replace("'", r"\'")),
                     test_files)
    return test_files


def mouse_scroll(driver, scroll_y):

    """scrolls by scroll_y in the netgraph div"""

    element = driver.find_element_by_id('netgraph')
    mouse_x = (element.location['x']+element.size['width'])/2.0
    mouse_y = (element.location['y']+element.size['height'])/2.0
    script = '''var netg = document.getElementById("netgraph");
                evt = document.createEvent("Event");
                evt.initEvent("wheel", true, true);
                evt.deltaY = %s;
                evt.clientX = %s;
                evt.clientY = %s ;
                netg.dispatchEvent(evt);'''
    driver.execute_script(script % (scroll_y, mouse_x, mouse_y))


def imgur_screenshot(driver):

    """Takes a screenshot, uploads it to imgur, and prints the link"""

    import pyimgur
    driver.get_screenshot_as_file('test_result.png')
    client_id = 'ce3e3bc9c9f0af0'
    client_secret = 'b033592e871bd14ac89d3e7356d8d96691713170'
    im = pyimgur.Imgur(client_id, client_secret)

    current_folder = os.getcwd()
    PATH = os.path.join(current_folder, 'test_result.png')
    uploaded_image = im.upload_image(PATH, title="Uploaded to Imgur")
    os.remove('test_result.png')
    print()
    print(uploaded_image.title)
    print(uploaded_image.link)



# Question Answering with Control

# This model shows a form of question answering where statements and 
# questions are supplied through a single 'visual input' and the replies 
# are produced in a 'motor output' as discussed in the book. You will 
# implement this by using the basal ganglia to store and retrieve 
# information from working memory in response to visual input. More 
# specifically, the basal ganglia decides what to do with the information 
# in the visual channel based on its content (i.e. whether it is a 
# statement or a question).

# **Note: Simplified method of building the model using the spa (semantic 
# pointer architecture) package in Nengo 2.0 is shown in the 
# spa_question-control.py file in the same folder. 

# This model has parameters as described in the book. Note that in Nengo 1.4, 
# network arrays were used to construct this model for computational reasons 
# as explained in the book. Nengo 2.0 has 'EnsembleArray' as an equivalent to 
# network arrays which you will use in this model.

# When you run the model, it will start by binding RED and CIRCLE and then 
# binding BLUE and SQUARE so the memory essentially has 
# RED * CIRCLE + BLUE * SQUARE. This is stored in memory because the model is 
# told that RED * CIRCLE is a STATEMENT (i.e. RED * CIRCLE + STATEMENT in the 
# code) as is BLUE * SQUARE. Then it is presented with something like 
# QUESTION + RED (i.e., "What is red?"). The basal ganglia then reroutes that 
# input to be compared to what is in working memory and the result shows up in 
# the motor channel.

# Press the play button to run the simulation.
# The graphs show that when the input to the Visual system is a STATEMENT, 
# there is no response from the Motor system and the input is stored in the 
# Memory. However, when the input to the Visual system is a QUESTION, the Motor 
# system responds with the appropriate answer. For instance, when the input to 
# Visual system is CIRCLE the output from the motor system is RED.


# Setup the environment
import numpy as np
import nengo
from nengo.spa import Vocabulary
from nengo.spa import BasalGanglia
from nengo.spa import Thalamus
from nengo.dists import Uniform

dim=100         # Number of dimensions 
N=30            # Neurons per dimension
N_conv=70       # Number of neurons per dimension in bind/unbind populations
N_mem=50        # Number of neurons per dimension in memory population
ZERO=[0]*dim    # Defining a zero vector having length equal to the number of 
                # dimensions

#Creating the vocabulary
rng = np.random.RandomState(15)
vocab = Vocabulary(dimensions=dim, rng=rng, max_similarity=0.05)

model = nengo.Network(label='Question Answering with Control', seed=15)
with model:
    #Ensembles
    visual = nengo.networks.EnsembleArray(n_neurons=N, n_ensembles=dim, 
                                max_rates=Uniform(100,300), label='Visual')
    channel = nengo.networks.EnsembleArray(n_neurons=N, n_ensembles=dim, 
                                label='Channel')
    motor = nengo.networks.EnsembleArray(n_neurons=N, n_ensembles=dim, 
                                label='Motor')   
    
    #Creating a memory (integrator)
    tau = 0.1
    memory = nengo.networks.EnsembleArray(n_neurons=N, n_ensembles=dim, 
                                label='Memory')
    nengo.Connection(memory.output, memory.input, synapse=tau)     
    
    #function for providing visual input
    def visual_input(t):
        if 0.1 < t < 0.3:
            return vocab.parse('STATEMENT+RED*CIRCLE').v
        elif 0.35 < t < 0.5:
            return vocab.parse('STATEMENT+BLUE*SQUARE').v
        elif 0.55 < t < 0.7:
            return vocab.parse('QUESTION+BLUE').v
        elif 0.75 < t < 0.9:
            return vocab.parse('QUESTION+CIRCLE').v
        else:
            return ZERO
        

    #function for flipping the output of the thalamus
    def xBiased(x):
        return [1-x]
     
    #Providing input to the model
    input = nengo.Node(output=visual_input, size_out=dim)
    nengo.Connection(input, visual.input)
    
    nengo.Connection(visual.output, channel.input, synapse=0.02)
    nengo.Connection(channel.output, memory.input)
    
    #Creating the unbind network
    unbind = nengo.networks.CircularConvolution(n_neurons=N_conv, 
                    dimensions=dim, invert_a=True)
    nengo.Connection(visual.output, unbind.A)
    nengo.Connection(memory.output, unbind.B)
    nengo.Connection(unbind.output, motor.input)
    
    #Creating the basal ganglia and the thalamus network
    BG = nengo.networks.BasalGanglia(dimensions=2)  
    thal = nengo.networks.Thalamus(dimensions=2)
    nengo.Connection(BG.output, thal.input, synapse=0.01)
    
    #Defining the transforms for connecting the visual input to the BG
    trans0 = np.matrix(vocab.parse('STATEMENT').v)
    trans1 = np.matrix(vocab.parse('QUESTION').v)
    nengo.Connection(visual.output, BG.input[0], transform=trans0)
    nengo.Connection(visual.output, BG.input[1], transform=trans1) 
    
    #Connecting thalamus output to the two gates gating the channel 
    # and the motor populations
    passthrough = nengo.Ensemble(n_neurons=N, dimensions=2)
    nengo.Connection(thal.output, passthrough)
 
    gate0 = nengo.Ensemble(N, 1, label='Gate0')
    nengo.Connection(passthrough[0], gate0, function=xBiased, synapse=0.01)    
    gate1 = nengo.Ensemble(N, 1, label='Gate1')
    nengo.Connection(passthrough[1], gate1, function=xBiased, synapse=0.01)
    
    for ensemble in channel.ea_ensembles:
        nengo.Connection(gate0, ensemble.neurons, 
                                transform=[[-3]] * gate0.n_neurons)
        
    for ensemble in motor.ea_ensembles:
        nengo.Connection(gate1, ensemble.neurons, 
                                transform=[[-3]] * gate1.n_neurons) 
      



# Tutorial 8: Combining Information

# Now that we can represent multiple things using the same group of neurons,
# we can also combine information together.  Here we introduce a new syntax
# when making Connections: just can specify which of the multiple dimensions
# to use.  For example, "Connection(a, c[0])" says to take the single dimension
# in a and pass it into the first dimension in c.  (In Python, we always start
# counting at zero, so the first dimension is x[0], then x[1], then x[2], and
# so on).  You can also do this when connecting out of a group, so you could do
# something like "Connection(c[1], d[3])".

# In the example below, we combine the information from two different Ensembles
# (a and b) into a third one (c) that represents both values.  Notice that this
# is different than the Addition tutorial in that we are keeping both values
# separate, rather than adding them together.

# Advanced Nengo Tip: Connections support full Python slice notation, so you
# can also do things like [-1], or [1:5] or [::-1] and so on.

import nengo

model = nengo.Network()
with model:
    stim_a = nengo.Node(0)
    a = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stim_a, a)
    
    stim_b = nengo.Node(0)
    b = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stim_b, b)
    
    c = nengo.Ensemble(n_neurons=200, dimensions=2)
    nengo.Connection(a, c[0])
    nengo.Connection(b, c[1])
    



# The simulations in these folders reproduce the tutorials from the book
# How to Build a Brain (Eliasmith, 2013, Oxford).  The original tutorials
# were written for Nengo 1.4.  The examples in these folders provide the final
# models for each of the tutorials in the relevant chapters, but implemented
# in Nengo 2.0.  
#
# You can follow the general form of each tutorial from the book,
# but there are significant differences in how models are built between Nengo
# versions.  The comment text in these simulation files describes those 
# differences as they relate to each example tutorial.  Consequently, you can
# just look at these files for all the information in the tutorials (except
# these files don't walk you through model construction).

import nengo
model = nengo.Network()


import struct
import collections

import nengo
import numpy as np

from nengo_gui.components.component import Component


class Raster(Component):
    """Plot showing spike events over time."""

    config_defaults = dict(n_neurons=10,
                           **Component.config_defaults)

    def __init__(self, obj):
        super(Raster, self).__init__()
        self.neuron_type = obj.neuron_type
        self.obj = obj.neurons
        self.data = collections.deque()
        self.max_neurons = obj.n_neurons

        self.conn = None
        self.node = None
        self.chosen = None



    def attach(self, page, config, uid):
        super(Raster, self).attach(page, config, uid)
        self.label = page.get_label(self.obj.ensemble)

    def add_nengo_objects(self, page):
        with page.model:
            self.node = nengo.Node(self.gather_data, size_in=self.max_neurons)
            if 'spikes' in self.neuron_type.probeable:
                self.conn = nengo.Connection(self.obj, self.node, synapse=None)

    def remove_nengo_objects(self, page):
        page.model.nodes.remove(self.node)
        if 'spikes' in self.neuron_type.probeable:
            page.model.connections.remove(self.conn)

    def gather_data(self, t, x):
        if self.chosen is None:
            self.compute_chosen_neurons()
        indices = np.nonzero(x[self.chosen])[0]
        data = struct.pack('<f%dH' % len(indices), t, *indices)
        self.data.append(data)

    def compute_chosen_neurons(self):
        n_neurons = self.page.config[self].n_neurons
        n_neurons = min(n_neurons, self.max_neurons)
        self.chosen = np.linspace(0, self.max_neurons-1, 
                                  n_neurons).astype(int)

    def update_client(self, client):
        while len(self.data) > 0:
            data = self.data.popleft()
            client.write_binary(data)

    def javascript(self):
        info = dict(uid=id(self), label=self.label,
                    max_neurons=self.max_neurons)
        json = self.javascript_config(info)
        return 'new Nengo.Raster(main, sim, %s);' % json

    def code_python_args(self, uids):
        return [uids[self.obj.ensemble]]

    def message(self, msg):
        if msg.startswith('n_neurons:'):
            n_neurons = min(int(msg[10:]), self.max_neurons)
            self.page.config[self].n_neurons = n_neurons
            self.compute_chosen_neurons()
            self.page.modified_config()



# Tutorial 16: Ensemble Properties

# In addition to the number of neurons, the number of dimensions, and the
# radius, there are other parameters that can be specified when creating an
# Ensemble.  Here are a few that may be useful.

# max_rates
#    Each neuron has a different maximum firing rate, and this parameter
#    specifies the random distribution controlling this.  The default is
#    a uniform distribution between 200Hz and 400Hz.

# encoders
#    Each neuron has a different preferred stimulus.  For a 1-dimensional
#    Ensemble, this means that half of the neurons prefer -1 and the other
#    half prefer +1.  This is why some neurons fire more for large values and
#    some for small values.  In the example below, we set all the encoders to
#    be +1.

# intercepts
#    Each neuron only starts firing when the similarity between the value and
#    its preferred value reaches a particular limit.  This is normally a
#    uniform distribution between -1 and 1.  Notice that if you set it to a
#    range like 0.5 to 1, then no neurons at all will fire for any values
#    between -0.5 and 0.5.  This means that any value between -0.5 and 0.5
#    will get reduced to exactly zero, which can be useful.

import nengo

model = nengo.Network()
with model:

    stim = nengo.Node(0)

    a = nengo.Ensemble(n_neurons=50, dimensions=1)

    b = nengo.Ensemble(n_neurons=50, dimensions=1,
                       max_rates=nengo.dists.Uniform(50, 100))

    c = nengo.Ensemble(n_neurons=50, dimensions=1,
                       encoders=nengo.dists.Choice([[1]]))

    d = nengo.Ensemble(n_neurons=50, dimensions=1,
                       intercepts=nengo.dists.Uniform(0.5, 1.0))

    nengo.Connection(stim, a)
    nengo.Connection(stim, b)
    nengo.Connection(stim, c)
    nengo.Connection(stim, d)



# Nengo Example: A Simple Harmonic Oscillator
#
# This demo implements a simple harmonic oscillator in a 2D neural 
# population.  The oscillator is more visually interesting on its own than 
# the integrator, but the principle at work is the same. Here, instead of 
# having the recurrent input just integrate (i.e. feeding the full input 
# value back to the population), we have two dimensions which interact. 

import nengo

model = nengo.Network()
with model:
    # Create the ensemble for the oscillator
    neurons = nengo.Ensemble(n_neurons=200, dimensions=2)

    # Create an input signal that gives a brief input pulse to start the
    # oscillator
    stim = nengo.Node(lambda t: 1 if t < 0.1 else 0)
    # Connect the input signal to the neural ensemble
    nengo.Connection(stim, neurons[0])

    # Create the feedback connection
    nengo.Connection(neurons, neurons, transform=[[1, 1], [-1, 1]], synapse=0.1)


import nengo

class NameFinder(object):
    def __init__(self, terms, net):
        self.base_terms = terms
        self.known_name = {}
        for k, v in terms.items():
            if not k.startswith('_'):
                try:
                    self.known_name[v] = k
                except TypeError:
                    pass
        self.find_names(net)

    def find_names(self, net):
        net_name = self.known_name[net]

        base_lists = ['ensembles', 'nodes', 'connections', 'networks']
        all_lists = ['all_ensembles', 'all_nodes', 'all_connections',
                     'all_networks', 'all_objects', 'all_probes']

        classes = (nengo.Node, nengo.Ensemble, nengo.Network,
                   nengo.Connection)

        for inst_attr in dir(net):
            private = inst_attr.startswith('_')
            in_lists = inst_attr in base_lists + all_lists
            if not private and not in_lists:
                attr = getattr(net, inst_attr)
                if isinstance(attr, list):
                    for i, obj in enumerate(attr):
                        if obj not in self.known_name:
                            n = '%s.%s[%d]' % (net_name, inst_attr, i)
                            self.known_name[obj] = n
                elif isinstance(attr, classes):
                    if attr not in self.known_name:
                        self.known_name[attr] = '%s.%s' % (net_name, inst_attr)


        for obj_type in base_lists:
            for i, obj in enumerate(getattr(net, obj_type)):
                name = self.known_name.get(obj, None)
                if name is None:
                    name = '%s.%s[%d]' % (net_name, obj_type, i)
                    self.known_name[obj] = name

        for n in net.networks:
            self.find_names(n)

    def name(self, obj):
        return self.known_name[obj]



# Arbitrary Linear Transformation

# This model shows that any linear transformation between ensembles can be
# easily computed by selecting an appropriate value for the "transform". It 
# provides an example of computing linear transformations using vector 
# representations.

# Network diagram:  [input - 2D] ---> (x - 2d) ---> (z - 3D)
# A two-dimensional input signal is first fed into a two-dimensional neuronal 
# ensemble x, which then projects it on to another three-dimensional ensemble z.

# This model contains the parameters as described in the book. Setting the 
# transform equal to the 'weight_matrix' is analogous to entering the weights 
# in the "2 to 3 Coupling Matrix" window in Nengo 1.4 GUI as described in the 
# book.

# Press the play button to run the simulation.
# The graphs show a two-dimesional input linearly projected on to a 
# two-dimensional ensemble of neurons (x), which further linearly projects it on
# to a three-dimesional neuronal ensemble (z). You can use the sliders to change
# the input values provided by the input node.

#Setup the environment
import numpy as np
import nengo

#Create a 'model' object to which we can add ensembles, connections, etc.  
model = nengo.Network(label="Arbitrary Linear Transformation")
with model:
    #Two-dimensional input signal with constant values of 0.5 and -0.5 in two 
    #dimensions
    input = nengo.Node(lambda t: [.5,-.5])
      
    #Ensembles with 200 LIF neurons having dimentions 2 and 3
    x = nengo.Ensemble(200, dimensions=2)
    z = nengo.Ensemble(200, dimensions=3)
       
    #Connect the input to ensemble x
    nengo.Connection(input, x)
    
    #Connect ensemble x to ensemble z using a weight matrix
    weight_matrix = [[0.0, 1.0], [1.0, 0.0], [0.5, 0.5]]   
    nengo.Connection(x, z, transform = weight_matrix)



"""Classes to instantiate and manage the life cycle of the nengo_gui
backend."""

from __future__ import print_function

import select
import signal
import sys
import threading
import webbrowser

import nengo_gui
from nengo_gui.guibackend import GuiServer


class ServerShutdown(Exception):
    """Causes the server to shutdown when raised."""
    pass


class BaseGUI(object):
    """Creates a basic nengo_gui backend server.

    Parameters
    ----------
    model_context : nengo_gui.backend.backend.ModelContext
        Model and its context served by the backend.
    server_settings : nengo_gui.backend.backend.GuiServerSettings, optional
        Backend settings.
    page_settings : nengo_gui.page.PageSettings, optional
        Frontend page settings.
    """
    def __init__(
            self, model_context, server_settings=None, page_settings=None):
        if server_settings is None:
            server_settings = nengo_gui.guibackend.GuiServerSettings()
        if page_settings is None:
            page_settings = nengo_gui.page.PageSettings()

        self.model_context = model_context

        self.server = GuiServer(
            self.model_context, server_settings, page_settings)

    def start(self):
        """Start the backend server and wait until it shuts down."""
        try:
            self.server.serve_forever(poll_interval=0.02)
        except ServerShutdown:
            self.server.shutdown()
        finally:
            self.server.wait_for_shutdown(0.05)


class InteractiveGUI(BaseGUI):
    """Creates a nengo_gui backend server and provides some useful information
    on stdout. Also registers handlers to allow a server shutdown with Ctrl-C.

    Parameters
    ----------
    model_context : nengo_gui.backend.backend.ModelContext
        Model and its context served by the backend.
    page_settings : nengo_gui.page.PageSettings, optional
        Frontend page settings.
    port : int
        Port to listen on.
    password : str, optional
        Password required to connect to the backend.
    """

    def start(self):
        protocol = 'https:' if self.server.settings.use_ssl else 'http:'
        print("Starting nengo server accessible at:\n  %s//%s:%d/?token=%s" % (
            protocol, 'localhost', self.server.server_port,
            self.server.auth_token))

        if not sys.platform.startswith('win'):
            signal.signal(signal.SIGINT, self._confirm_shutdown)

        try:
            self.server.serve_forever(poll_interval=0.02)
            print("No connections remaining to the nengo_gui server.")
        except ServerShutdown:
            self.server.shutdown()
        finally:
            print("Shutting down server...")

            self.server.wait_for_shutdown(0.05)
            n_zombie = sum(thread.is_alive()
                           for thread, _ in self.server.requests)
            if n_zombie > 0:
                print("%d zombie threads will close abruptly" % n_zombie)

    def _confirm_shutdown(self, signum, frame):
        signal.signal(signal.SIGINT, self._immediate_shutdown)
        sys.stdout.write("\nShut-down this web server (y/[n])? ")
        sys.stdout.flush()
        rlist, _, _ = select.select([sys.stdin], [], [], 10)
        if rlist:
            line = sys.stdin.readline()
            if line[0].lower() == 'y':
                raise ServerShutdown()
            else:
                print("Resuming...")
        else:
            print("No confirmation received. Resuming...")
        signal.signal(signal.SIGINT, self._confirm_shutdown)

    def _immediate_shutdown(self, signum, frame):
        raise ServerShutdown()


class GUI(InteractiveGUI):
    """Starts an InteractiveGUI.

    Provides an easier instantiation syntax for the use in scripts.

    Parameters
    ----------
    filename : str
        Filename of the calling script (usually you want to use ``__file__``).
    model : nengo.Network
        The model to visualize.  If None, the script filename is evaluated to
        create the model.
    locals : dict
        The locals() dictionary after running the script.  If None, it is
        determinined by running the script filename.  Its contents are used
        to give useful names for the objects in the model.
    editor : bool
        Whether or not to show the editor
    """
    def __init__(self, filename=None, model=None, locals=None, editor=True):
        if not editor:
            ps = nengo_gui.page.PageSettings(
                    editor_class=nengo_gui.components.editor.NoEditor)
        else:
            ps = nengo_gui.page.PageSettings()
        super(GUI, self).__init__(nengo_gui.guibackend.ModelContext(
            filename=filename, model=model, locals=locals),
            page_settings=ps)

    def start(self):
        t = threading.Thread(
            target=webbrowser.open,
            args=('http://localhost:%d/?token=%s' % (
                self.server.server_port,
                self.server.settings.auth_token,)))
        t.start()

        super(GUI, self).start()



#!/usr/bin/env python
#
# This code is part of Grandalf
# Copyright (C) 2008-2011 Axel Tillequin (bdcht3@gmail.com)
# published under GPLv2 license or EPLv1 license

from .utils import Poset
from operator import attrgetter

#------------------------------------------------------------------------------
#  vertex_core class:
#  e: list of edge_core objects.
#------------------------------------------------------------------------------
class  vertex_core(object):
    def __init__(self):
        # will hold list of edges for this vertex (adjacency list)
        self.e = []

    def deg(self): return len(self.e)

    def e_in(self):
        return filter( (lambda e:e.v[1]==self), self.e )

    def e_out(self):
        return filter( (lambda e:e.v[0]==self), self.e )

    def e_dir(self,dir):
        if dir>0: return self.e_out()
        if dir<0: return self.e_in()
        return self.e

    def N(self,f_io=0):
        N = []
        if f_io<=0: N += [ e.v[0] for e in self.e_in()  ]
        if f_io>=0: N += [ e.v[1] for e in self.e_out() ]
        return N

    def e_to(self,y):
        for e in self.e_out():
            if e.v[1]==y: return e
        return None

    def e_from(self,x):
        for e in self.e_in():
            if e.v[0]==x: return e
        return None

    def e_with(self,v):
        for e in self.e:
            if v in e.v: return e
        return None

    def detach(self):
        E = self.e[:]
        for e in E: e.detach()
        assert self.deg()==0
        return E


#------------------------------------------------------------------------------
#  edge_core class:
#  v = (x,y): a tuple of vertices (probably of type Vertex, see below)
#------------------------------------------------------------------------------
class  edge_core(object):
    def __init__(self,x,y):
        self.deg = 0 if x==y else 1
        self.v = (x,y)


#------------------------------------------------------------------------------
#  Vertex class enhancing a vertex_core with
#  c: the poset of connected vertices containing this vertex.
#  data: anything else associated with the vertex.
#------------------------------------------------------------------------------
class  Vertex(vertex_core):
    counter=0
    def __init__(self,data=None):
        vertex_core.__init__(self)
        self.index=Vertex.counter
        Vertex.counter += 1
        # by default, a new vertex belongs to its own component
        # but when the vertex is added to a graph, c points to the
        # connected component where it belongs.
        self.c = None
        self.data = data
    @classmethod
    def count(cls):
        return cls.counter
    def __hash__(self):
        return self.index

#------------------------------------------------------------------------------
#  Edge class:
#  w: weight associated with the edge, defaults to 1.
#  data: anything else associated with the edge.
#------------------------------------------------------------------------------
class  Edge(edge_core):
    counter=0
    def __init__(self,x,y,w=1,data=None,connect=False):
        edge_core.__init__(self,x,y)
        self.index=Edge.counter
        Edge.counter += 1
        # w is an optional weight associated with the edge.
        self.w = w
        self.data = data
        self.feedback = False
        if connect and (x.c==None or y.c==None):
            c = x.c or y.c
            c.add_edge(self)

    @classmethod
    def count(cls):
        return cls.counter
    def __hash__(self):
        return self.index

    def attach(self):
        if not self in self.v[0].e : self.v[0].e.append(self)
        if not self in self.v[1].e : self.v[1].e.append(self)

    def detach(self):
        if self.deg==1:
            assert self in self.v[0].e
            assert self in self.v[1].e
            self.v[0].e.remove(self)
            self.v[1].e.remove(self)
        else:
            if self in self.v[0].e: self.v[0].e.remove(self)
            assert self not in self.v[0].e
        return [self]


#------------------------------------------------------------------------------
#  graph_core class: A connected graph of Vertex/Edge objects.
#  self.sV: set of vertices
#  self.sE: set of edges
#  The graph is stored in edge list representation by self.sE,
#  but since the vertex_core embbeds edges information, the adjacency list rep
#  is straightforward from self.sV.
#------------------------------------------------------------------------------
class  graph_core(object):
    def __init__(self,V=None,E=None,directed=True):

        if V is None: V=[]
        if E is None: E=[]

        self.directed = directed

        self.sV = Poset(V)
        self.sE = Poset([])

        self.degenerated_edges=[]

        if len(self.sV)==1:
            v = self.sV.o[0]
            v.c = self
            for e in v.e: e.detach()
            return

        for e in E:
            x = e.v[0]
            y = e.v[1]
            if (x in V) and (y in V):
                if e.deg==0:
                    e.detach()
                    self.degenerated_edges.append(e)
                    continue
                e.attach()
                self.sE.add(e)
                if x.c is None: x.c=Poset([x])
                if y.c is None: y.c=Poset([y])
                if id(x.c)!=id(y.c):
                    x.c.update(y.c)
                    y.c=x.c
                s=x.c
            else:
                raise ValueError('unknown Vertex (%s or %s)'%(x.data,y.data))
        #check if graph is connected:
        for v in self.V():
            if v.c is None or (v.c!=s):
                raise ValueError('unconnected Vertex %s'%v.data)
            else:
                v.c = self

    # allow a graph_core to hold a single vertex:
    def add_single_vertex(self,v):
        if len(self.sE)==0 and len(self.sV)==0:
            self.sV.add(v)
            v.c = self

    # add edge e. At least one of its vertex must belong to the graph,
    # the other being added automatically.
    def add_edge(self,e):
        if e not in self.sE:
            x = e.v[0]
            y = e.v[1]
            if not ((x in self.sV) or (y in self.sV)):
                raise ValueError('unconnected edge')
            self.sV.add(x)
            self.sV.add(y)
            e.attach()
            self.sE.add(e)
            x.c = self
            y.c = self

    # remove Edge :
    # this procedure checks that the resulting graph is connex.
    def remove_edge(self,e):
        if e.deg==0: return
        e.detach()
        # check if still connected (path is not oriented here):
        if not self.path(e.v[0],e.v[1]):
            # return to inital state by reconnecting everything:
            e.attach()
            # exit with exception!
            raise ValueError(e)
        else:
            self.sE.remove(e)

    # remove Vertex:
    # this procedure checks that the resulting graph is connex.
    def remove_vertex(self,x):
        V = x.N() #get all neighbor vertices to check paths
        E = x.detach() #remove the edges from x and neighbors list
        # now we need to check if all neighbors are still connected,
        # and it is sufficient to check if one of them is connected to
        # all others:
        v0 = V.pop(0)
        for v in V:
            if not self.path(v0,v):
                # repair everything and raise exception if not connected:
                for e in E: e.attach()
                raise ValueError(x)
        # remove edges and vertex from internal sets:
        for e in E: self.sE.remove(e)
        self.sV.remove(x)
        x.c = None

    # generates an iterator over vertices, with optional filter
    def V(self,cond=None):
        V = self.sV
        if cond is None: cond=(lambda x:True)
        for v in V:
            if cond(v):
                yield v

    # generates an iterator over edges, with optional filter
    def E(self,cond=None):
        E = self.sE
        if cond is None: cond=(lambda x:True)
        for e in E:
            if cond(e):
                yield e

    # vertex/edge properties :
    #-------------------------
    # returns number of vertices
    def order(self):
        return len(self.sV)

    # returns number of edges
    def norm(self):
        return len(self.sE)

    # returns the minimum degree
    def deg_min(self):
        return min(map(vertex_core.deg,self.sV))

    # returns the maximum degree
    def deg_max(self):
        return max(map(vertex_core.deg,self.sV))

    # returns the average degree d(G)
    def deg_avg(self):
        return sum(map(vertex_core.deg,self.sV))/float(self.order())

    # returns the epsilon value (number of edges of G per vertex)
    def eps(self):
        return float(self.norm())/self.order()

    # shortest path between vertices x and y by breadth-first descent
    def path(self,x,y,f_io=0,hook=None):
        assert x in self.sV
        assert y in self.sV
        if x==y: return []
        if f_io!=0: assert self.directed==True
        # path:
        p = None
        if hook is None: hook = lambda x:False
        # apply hook:
        hook(x)
        # visisted:
        v = {x:None}
        # queue:
        q = [x]
        while (not p) and len(q)>0:
            c = q.pop(0)
            for n in c.N(f_io):
                if not v.has_key(n):
                    hook(n)
                    v[n] = c
                    if n==y: p = [n]
                    q.append(n)
                if p: break
        #now we fill the path p backward from y to x:
        while p and p[0]!=x:
            p.insert(0,v[p[0]])
        return p

    # shortest weighted-edges paths between x and all other vertices
    # by dijkstra's algorithm with heap used as priority queue.
    def dijkstra(self,x,f_io=0,hook=None):
        from collections import defaultdict
        from heapq import heappop, heappush
        if x not in self.sV: return None
        if f_io!=0: assert self.directed==True
        # take a shallow copy of the set of vertices containing those for which
        # the shortest path to x needs to be computed:
        S = self.sV.copy()
        # initiate with path to itself...
        v = x
        # D is the returned vector of distances:
        D = defaultdict(lambda :None)
        D[v] = 0.0
        L = [(D[v],v)]
        while len(S)>0:
            #pdb.set_trace()
            l,u = heappop(L)
            S.remove(u)
            for e in u.e:
                v = e.v[0] if (u is e.v[1]) else e.v[1]
                Dv = l+e.w
                if D[v]!=None:
                    # check if heap/D needs updating:
                    # ignore if a shorter path was found already...
                    if Dv<D[v]:
                        for i,t in enumerate(L):
                            if t[1] is v:
                                L.pop(i)
                                break
                        D[v]=Dv
                        heappush(L,(Dv,v))
                else:
                    D[v]=Dv
                    heappush(L,(Dv,v))
        return D

    # returns the set of strongly connected components
    # ("scs") by using Tarjan algorithm.
    # These are maximal sets of vertices such that there is a path from each
    # vertex to every other vertex.
    # The algorithm performs a DFS from the provided list of root vertices.
    # A cycle is of course a strongly connected component,
    # but a strongly connected component can include several cycles.
    # The Feedback Acyclic Set of edge to be removed/reversed is provided by
    # marking the edges with a "feedback" flag.
    # Complexity is O(V+E).
    def get_scs_with_feedback(self,roots):
        from  sys import getrecursionlimit,setrecursionlimit
        limit=getrecursionlimit()
        N=self.norm()+10
        if N>limit:
            setrecursionlimit(N)
        def _visit(v,L):
            v.ind = v.ncur
            v.lowlink = v.ncur
            Vertex.ncur += 1
            self.tstack.append(v)
            v.mark = True
            for e in v.e_out():
                w = e.v[1]
                if w.ind==0:
                    _visit(w,L)
                    v.lowlink = min(v.lowlink,w.lowlink)
                elif w.mark:
                    e.feedback = True
                if w in self.tstack:
                    v.lowlink = min(v.lowlink,w.ind)
            if v.lowlink==v.ind:
                l=[self.tstack.pop()]
                while l[0]!=v:
                    l.insert(0,self.tstack.pop())
                #print "unstacked %s"%('-'.join([x.data[1:13] for x in l]))
                L.append(l)
            v.mark=False
        self.tstack=[]
        scs = []
        Vertex.ncur=1
        for v in self.sV: v.ind=0
        # start exploring tree from roots:
        for v in roots:
            if v.ind==0: _visit(v,scs)
        # now possibly unvisited vertices:
        for v in self.sV:
            if v.ind==0: _visit(v,scs)
        # clean up Tarjan-specific data:
        for v in self.sV:
            del v.ind
            del v.lowlink
            del v.mark
        del Vertex.ncur
        del self.tstack
        setrecursionlimit(limit)
        return scs

    # returns neighbours of a vertex v:
    # f_io=-1 : parent nodes
    # f_io=+1 : child nodes
    # f_io= 0 : all (default)
    def N(self,v,f_io=0):
        return v.N(f_io)

    # general graph properties:
    # -------------------------

    # returns True iff
    #  - o is a subgraph of self, or
    #  - o is a vertex in self, or
    #  - o is an edge in self
    def __contains__(self,o):
        try:
            return o.sV.issubset(self.sV) and o.sE.issubset(self.sE)
        except AttributeError:
            return ((o in self.sV) or (o in self.sE))

    # merge graph_core G into self
    def union_update(self,G):
        for v in G.sV: v.c = self
        self.sV.update(G.sV)
        self.sE.update(G.sE)

    # derivated graphs:
    # -----------------

    # returns subgraph spanned by vertices V
    def spans(self,V):
        raise NotImplementedError

    # returns join of G (if disjoint)
    def __mul__(self,G):
        raise NotImplementedError

    # returns complement of a graph G
    def complement(self,G):
        raise NotImplementedError

    # contraction G\e
    def contract(self,e):
        raise NotImplementedError


#------------------------------------------------------------------------------
#  Graph class: Disjoint-set Graph.
#  V: list/set of vertices of type Vertex.
#  E: list/set of edges of type Edge.
#  The graph is stored in disjoint-sets holding each connex component
#  in self.C as a list of graph_core objects.
#------------------------------------------------------------------------------
class  Graph(object):
    def __init__(self,V=None,E=None,directed=True):
        if V is None: V=[]
        if E is None: E=[]
        self.directed = directed
        # tag connex set of vertices:
        # at first, every vertex is its own component
        for v in V: v.c = Poset([v])
        CV = [v.c for v in V]
        # then pass through edges and union associated vertices such that
        # CV finally holds only connected sets:
        for e in E:
            x = e.v[0]
            y = e.v[1]
            assert x in V
            assert y in V
            assert x.c in CV
            assert y.c in CV
            e.attach()
            if x.c!=y.c:
                #merge y.c into x.c :
                x.c.update(y.c)
                #update set list (MUST BE DONE BEFORE UPDATING REFS!)
                CV.remove(y.c)
                #update reference:
                for z in y.c: z.c = x.c
        # now create edge sets from connected vertex sets and
        # make the graph_core connected graphs for this component :
        self.C = []
        for c in CV:
            s = set()
            for v in c: s.update(v.e)
            self.C.append(graph_core(c,s,directed))

    # add vertex v into the Graph as a new (unconnected) component
    def add_vertex(self,v):
        for c in self.C:
            if (v in c.sV): return 0
        g = graph_core(directed=self.directed)
        g.add_single_vertex(v)
        self.C.append(g)
        return 1

    # add edge e and its vertices into the Graph possibly merging the
    # associated graph_core components
    def add_edge(self,e):
        # take vertices:
        x = e.v[0]
        y = e.v[1]
        self.add_vertex(x)
        self.add_vertex(y)
        # take respective graph_cores:
        cx = x.c
        cy = y.c
        # add edge:
        cy.add_edge(e)
        # connect (union) the graphs:
        if cx!=cy:
            cx.union_update(cy)
            self.C.remove(cy)

    def get_vertices_count(self):
        total = 0
        for c in self.C:
            V = c.sV
            total += len(V)
        return total

    # generates an iterator over vertices
    def V(self):
        for c in self.C:
            V = c.sV
            for v in V: yield v

    # generates an iterator over edges
    def E(self):
        for c in self.C:
            E = c.sE
            for e in E: yield e

    # remove Edge from a core
    def remove_edge(self,e):
        # get the graph_core:
        c = e.v[0].c
        assert c==e.v[1].c
        assert c in self.C
        # remove edge in graph_core and replace it with two new cores
        # if removing edge disconnects the graph_core:
        try:
            c.remove_edge(e)
        except ValueError:
            e.detach()
            c.sE.remove(e)
            self.C.remove(c)
            tmpg = Graph(c.sV,c.sE,self.directed)
            assert len(tmpg.C)==2
            self.C.extend(tmpg.C)

    # remove a Vertex and all its edges from a core
    def remove_vertex(self,x):
        # get the graph_core:
        c = x.c
        assert c in self.C
        try:
            c.remove_vertex(x)
            if c.order()==0: self.C.remove(c)
        except ValueError:
            for e in x.detach(): c.sE.remove(e)
            c.sV.remove(x)
            self.C.remove(c)
            tmpg = Graph(c.sV,c.sE,self.directed)
            assert len(tmpg.C)==2
            self.C.extend(tmpg.C)

    # vertex/edge properties :
    #-------------------------

    def order(self):
        return sum(map(graph_core.order,self.C))

    def norm(self):
        return sum(map(graph_core.norm,self.C))

    def deg_min(self):
        return min(map(graph_core.deg_min,self.C))

    def deg_max(self):
        return max(map(graph_core.deg_max,self.C))

    def deg_avg(self):
        t = 0.0
        for c in self.C: t += sum(map(vertex_core.deg,c.sV))
        return t/self.order()

    def eps(self):
        return float(self.norm())/self.order()

    def path(self,x,y,f_io=0,hook=None):
        if x==y: return []
        if x.c!=y.c: return None
        # path:
        return x.c.path(x,y,f_io,hook)

    def N(self,v,f_io=0):
        return v.N(f_io)

    def __contains__(self,G):
        r = False
        for c in self.C: r |= (G in c)
        return r

    # returns True if Graph is connected
    def connected(self):
        return len(self.C)==1

    # returns connectivity (kappa)
    def connectivity(self):
        raise NotImplementedError

    # returns edge-connectivity (lambda)
    def e_connectivity(self):
        raise NotImplementedError

    # returns the list of graphs components
    def components(self):
        return self.C

    # derivated graphs:
    # -----------------

    # returns subgraph spanned by vertices V
    def spans(self,V):
        raise NotImplementedError

    # returns join of G (if disjoint)
    def __mul__(self,G):
        raise NotImplementedError

    # returns complement of a graph G
    def complement(self,G):
        raise NotImplementedError

    # contraction G\e
    def contract(self,e):
        raise NotImplementedError




# Communication Channel

# This example demonstrates how to create a connections from one neuronal
# ensemble to another that behaves like a communication channel (that is, it
# transmits information without changing it).

import numpy as np
import nengo

model = nengo.Network()
with model:
    # Create an abstract input signal that oscillates as sin(t)
    sin = nengo.Node(np.sin)

    # Create the neuronal ensembles
    a = nengo.Ensemble(n_neurons=100, dimensions=1)
    b = nengo.Ensemble(n_neurons=100, dimensions=1)

    # Connect the input to the first neuronal ensemble
    nengo.Connection(sin, a)

    # Connect the first neuronal ensemble to the second using a
    # neurotransmitter with a 10ms time constant
    # This is the communication channel.
    nengo.Connection(a, b, synapse=0.01)



import nengo
import numpy as np

model = nengo.Network()
with model:
    a = nengo.Ensemble(n_neurons=100, dimensions=1,
                       neuron_type=nengo.LIF(tau_rc=0.02, tau_ref=0.002))
    stim = nengo.Node(0)
    nengo.Connection(stim, a)
    
    output = nengo.Node(None, size_in=1)
    
    def my_function(x):
        if x<0:
            return -1
        else:
            return 1

    nengo.Connection(a, output, synapse=0.005,
                     function=my_function)
    
    
